import sys
from typing import Collection, Literal

_MAX_CANDIDATE_ITEMS = 750
_MAX_STRING_SIZE = 40
_MOVE_COST = 1
_CASE_COST = 1

VarsGroup = Literal["UC", "UP", "RC", "RP", "DU"]


_SUBSTITUTE_COST = 1
_TRANSPOSE_COST = 1


def osa_distance(a, b, max_cost=sys.maxsize):
    """Compute the Optimal String Alignment (OSA) distance between strings a and b.

    This is similar to Damerau-Levenshtein but doesn't allow multiple edits on substrings.
    """
    # Both strings are the same
    if a == b:
        return 0

    # Trim away common affixes
    pre = 0
    while a[pre:] and b[pre:] and a[pre] == b[pre]:
        pre += 1
    a = a[pre:]
    b = b[pre:]
    post = 0
    while a[:post or None] and b[:post or None] and a[post - 1] == b[post - 1]:
        post -= 1
    a = a[:post or None]
    b = b[:post or None]
    if not a or not b:
        return min(_MOVE_COST * (len(a) + len(b)), max_cost + 1)
    if len(a) > _MAX_STRING_SIZE or len(b) > _MAX_STRING_SIZE:
        return max_cost + 1

    # Initialize distance matrix
    d = [[0] * (len(b) + 1) for _ in range(len(a) + 1)]

    # Initialize first row and column
    for i in range(len(a) + 1):
        d[i][0] = i * _MOVE_COST
    for j in range(len(b) + 1):
        d[0][j] = j * _MOVE_COST

    # Fill the distance matrix
    for i in range(1, len(a) + 1):
        for j in range(1, len(b) + 1):
            cost = 0 if a[i - 1] == b[j - 1] else _SUBSTITUTE_COST

            d[i][j] = min(
                d[i - 1][j] + _MOVE_COST,  # deletion
                d[i][j - 1] + _MOVE_COST,  # insertion
                d[i - 1][j - 1] + cost  # substitution
            )

            # Check for transposition
            if (i > 1 and j > 1 and
                    a[i - 1] == b[j - 2] and
                    a[i - 2] == b[j - 1]):
                d[i][j] = min(
                    d[i][j],
                    d[i - 2][j - 2] + _TRANSPOSE_COST
                )

            # Early exit if we've exceeded max_cost
            if d[i][j] > max_cost:
                return max_cost + 1

    return min(d[len(a)][len(b)], max_cost + 1)


def find_closest_matches(wrong_name: str,
                         candidates: Collection[str],
                         max_items: int=_MAX_CANDIDATE_ITEMS,
                         max_distance: int=None) -> list[tuple[int, str]]:
    """从候选词中找出与目标词最接近的匹配项

    Args:
        wrong_name: 要匹配的目标词
        candidates: 候选词列表
        max_items: 最大候选词数量限制
        max_distance: 最大允许的编辑距离(为None时使用1/3名称长度)

    Returns:
        一个按距离排序的列表
        例如：[(1, 'typo'), (3, 'typing')]
    """
    if len(candidates) > max_items:
        return []

    wrong_name_len = len(wrong_name)
    if wrong_name_len > _MAX_STRING_SIZE:
        return []

    if max_distance is None:
        # 默认计算最大允许距离
        max_distance = (len(wrong_name) + 3) * _MOVE_COST // 3

    matches = []
    for candidate in candidates:
        if candidate == wrong_name:
            continue

        distance = osa_distance(wrong_name, candidate, max_distance)
        if distance <= max_distance:
            matches.append((distance, candidate))

    # 按距离排序并返回
    return sorted(matches, key=lambda x: x[0])


def merge_sorted_suggestions(suggestions_groups: dict[str, list[tuple[int, str]]],
                             if_only_group: str=None) -> tuple[list[tuple[int, str]], list[tuple[int, str]]]:
    """合并多个有序建议列表

    Args:
        suggestions_groups: 多个已排序的建议列表
        if_only_group: 如果存在，则将本组和其他组的建议分离返回

    Returns:
        if-only组列表与合并后的有序建议列表
    """
    import heapq

    # 将每个列表转换为迭代器
    iters = [iter(group) for type_, group in suggestions_groups.items() if if_only_group != type_]
    # 初始化堆，取每个列表的第一个元素
    heap = []
    for i, it in enumerate(iters):
        try:
            item = next(it)
            heapq.heappush(heap, (item[0], i, item[1]))
        except StopIteration:
            pass

    # 合并排序
    merged_suggestions = []
    while heap:
        score, i, name = heapq.heappop(heap)
        merged_suggestions.append((score, name))
        try:
            next_item = next(iters[i])
            heapq.heappush(heap, (next_item[0], i, next_item[1]))
        except StopIteration:
            pass

    if if_only_group:
        return suggestions_groups[if_only_group], merged_suggestions
    else:
        return [], merged_suggestions


def find_weighted_closest_matches(wrong_name: str,
                                  candidate_vars: dict[VarsGroup, Collection[str]]) -> dict[str, list[tuple[int, str]]]:
    """从候选变量中找出与目标变量最接近的匹配项
    Args:
        wrong_name: 要匹配的目标变量名
        candidate_vars: 候选变量列表
    Returns:
        与目标变量同类别的候选变量列表和其他合并后的有序建议列表
    """
    suggestions_groups: dict[str, list[tuple[int, str]]] = {}
    for candidate_type, candidates in candidate_vars.items():
        suggestions = find_closest_matches(wrong_name, candidates)
        suggestions_groups[candidate_type] = suggestions
    return suggestions_groups

if __name__ == "__main__":
    # sqr 与 sqrt 的距离
    # sqr 与 sin 的距离
    print(osa_distance("sqr", "sqrt")) # 2
    print(osa_distance("sqr", "sin")) # 2

    # 为什么都是2？为什么插入消费是2？




"""
这里定义错误处理器的基类，动态扩展的示例将在这里展示。
"""
import linecache
from types import TracebackType
from typing import Generator, Any, Type, final

import astroid

from kawaiitb.kraceback import KTBException, FrameSummary
from kawaiitb.runtimeconfig import rc
from kawaiitb.utils.ast_parse import astroid_walk_inside

__all__ = [
    "ErrorSuggestHandler"
]

@KTBException.register
class ErrorSuggestHandler:
    """
    异常处理器的基类。

    优先级最高的处理器会最先认领异常。
    在KTBException初始化时会顺带初始化所有的处理器。
    在格式化时，KTBException传入自身并判断各处理器是否能处理异常，然后选择优先级最高的处理器。

    优先级原则：
    - 仅ErrorSuggestHandler父类使用0.0优先级。低于此优先级的均不可能认领。
    - 原生级处理器使用1.0优先级。即本文件中的所有附加处理器。
    - 希望官方处理器先认领，认领失败时自己处理时，使用属于(0, 1)的优先级。
    - 希望自己的处理器先认领，认领失败时官方处理器处理时，使用属于(1, +infny)的优先级。
        - 独立的异常类型，使用优先级2
          (如ModuleNotFoundErrorHandler专注于ModuleNotFoundError)。
        - 独立的异常描述级别的，使用优先级3
          (如NoneTypeErrorHandler处理参数为None的TypeError)。
        - 针对特定异常参数/信息的建议/翻译，使用优先级4
          (如PyyamlNotFoundErrorHandler处理pyyaml的NameError)。
    """

    __priority__: float = 0.0  # 基处理器的优先级为0.0
    # 所有有效的处理器都应该高于此优先级以覆盖处理。
    # 所有不生效(如仅翻译)的处理器都应该低于此优先级。建议使用标准的: -1.0.

    def __init__(self, exc_type: Type[BaseException], exc_value: BaseException, exc_traceback: TracebackType, *, limit=None,
                 lookup_lines=True, capture_locals=False, compact=False,
                 max_group_width=15, max_group_depth=10, _seen=None):
        ...
        # 此处可以添加一些初始化逻辑，比如确定自己能不能处理这个异常。
        # 也可以根据这个异常帧的信息处理一些值。
        # 上面传入的东西就是一个异常发生时所有传给你的信息，包括异常的语句，上下文代码，这这那那的。
        # 可以看一些预设的处理器了解一下如何处理这些信息。

        # 另外，如果你是直接继承的ErrorSuggestHandler，其实可以省略super调用，
        # 因为init实际上并没有做什么……

        # 如果也不需要后面的设置，可以用**kwargs来收取所有参数，然后按需提取和忽略。、

    def __init_subclass__(cls, priority):
        """
        初始化子类时，自动注册翻译键。
        可以通过priority参数来设置处理器的优先级。
>>> class MyHandler(ErrorSuggestHandler, priority=2.0):
>>>     ...
        """
        cls.__priority__ = priority

    @property
    def priority(self) -> float:
        return self.__priority__

    def can_handle(self, ktb_exc: KTBException) -> bool:
        """返回本处理器是否能处理异常。"""
        # 处理器接受异常需要满足以下条件：
        # 1. 处理器声明自己能处理异常。
        # 2. 没有更高优先级的处理器能处理异常。
        # 举例来说，ModuleNotFoundErrorHandler和PyyamlNotFoundErrorHandler都能处理`import pyyaml`
        # 但前者只是基本的找不到包提醒，后者则是提供了更详细的解决方案，提示用户应该导入的是yaml。
        # 所以优先级PyyamlNotFoundErrorHandler(4.0) > ModuleNotFoundErrorHandler(2.0)。

        return True  # 当然这里是基处理器，要处理所有异常

    @classmethod
    def translation_keys(cls) -> dict[str, dict[str, Any]]:
        """
        翻译键。
        每个处理器都可以有自己的翻译键，建议非官方逻辑的处理器使用"exthandler.<namespace>.xxx"命名翻译键。
        返回格式是标准的静态文件格式，即{<language>: {<key>: <value>}}。
        **如果非空，则必须有"default"键**。

        如果你希望单纯翻译一个自定义处理器，你可以定义一个优先级为-1的处理器，然后在其中定义翻译键并注册。
        """
        ...

    def handle(self, ktb_exc: KTBException) -> Generator[str, None, None]:
        """
        处理异常，返回一个生成器，生成器会逐行产生处理后的错误信息。
        并不一定非要每行一定断一下，这只是为了模块的灵活性，如果你需要输出确定不变的多行信息，直接写就是了。
        """
        stype = ktb_exc.exc_type.__qualname__
        smod = ktb_exc.exc_type.__module__
        if smod not in ("__main__", "builtins"):
            if not isinstance(smod, str):
                smod = "<unknown>"
            stype = smod + '.' + stype
        yield rc.exc_line(stype, ktb_exc.final_exc_str)

    @staticmethod
    @final
    def parse_ast_from_exc(exc_frame: FrameSummary, parse_line=False):
        """
        从一个帧中解析并产生相关的AST节点
        """
        start_line = exc_frame.lineno
        end_line = exc_frame.end_lineno
        start_col = 0 if parse_line else exc_frame.colno
        end_col = 99999999 if parse_line else exc_frame.end_colno
        tree = astroid.parse("".join(linecache.getlines(exc_frame.filename)))

        yield from astroid_walk_inside(tree, start_line, end_line, start_col, end_col)




# 以下是示例代码
class ImportErrorHandler(ErrorSuggestHandler, priority=2.0):
    ...

#@KTBException.register  # 使用装饰器可以注册处理器。本示例中暂不使用。
class PyYamlImportErrorHandler(ImportErrorHandler, priority=4.0):
    """
    在pyyaml导入失败时提供更详细的解决方案。

    一个Handler将在KTBException初始化时被初始化
    初始化时，__init__会被传入异常的所有原始参数
    (不需要的参数可以用**kwargs原样丢回super)
    之后，KTBException会调用can_handle来判断是否能处理异常
    如果最终以最高优先级被选中，则会调用handle来处理异常
    """
    def __init__(self, exc_type, exc_value, exc_traceback, **kwargs):
        super().__init__(exc_type, exc_value, exc_traceback, **kwargs)

        # 检查是否是pyyaml相关的导入错误
        # 这里从__init__接受的参数的一部分也可以在后面从KTBException中获取,
        # 但其中的一些参数可能会在处理时被修改，所以最好还是从这里获取。
        # 如果需要修改后的参数，可以在handle中获取。

        self._can_handle = (issubclass(exc_type, ImportError) and
                            getattr(exc_value, "name", "") == "pyyaml")

    # 注册扩展翻译键到运行时配置，以供后续使用
    @classmethod
    def translation_keys(cls):
        return {
            "default": {
                # 没必要分行yield，直接写多行文本就好了~！
                "exthandler.pyyaml.hint": "Attention: You may have installed the 'pyyaml' package, but you should import 'yaml' instead of 'pyyaml'.\n"
                                          "- Install pyyaml using pip: pip install pyyaml\n"
                                          "- Import yaml in your code: import yaml",
            },
            "zh_hans": {
                "exthandler.pyyaml.hint": "注意：你可能已经安装了pyyaml包，但导入时应该使用'yaml'而不是'pyyaml'。\n"
                                          "- 使用pip安装pyyaml: pip install pyyaml\n"
                                          "- 在代码中导入yaml: import yaml",
            }
        }

    def can_handle(self, ktb_exc) -> bool:
        # 因为__init__和can_handle都是每个handle的必经之路
        # 所以在那边存储在这处理和直接在那处理保存结果效率都差不多
        return self._can_handle

    def handle(self, ktb_exc) -> Generator[str, None, None]:
        """
        1. 继承原则
        在本测试用例PyYamlImportErrorHandler中,
        如果直接继承 ErrorSuggestHandler 会导致
        super处理时直接把"ImportError: No module named 'pyyaml'"
        这条原生信息直接丢给用户. 而 ImportErrorHandler一定会为此准备。
        即使你不打算使用 ImportErrorHandler 提供的信息
        我仍然建议你继承 ImportErrorHandler, 这可以使得逻辑更清晰
        也可以避免一些潜在问题.
        2. 翻译原则
        你可以使用 rc.translate 来获取翻译, 也可以直接硬编码.
        但我强烈建议你使用翻译, 并至少准备default的英语语种.
        这样你的代码可以令全世界程序员都能看懂.
        3. yield原则
        如上, __init__和can_handle都是每个handle的必经之路,
        但只有handle是处理器确定要处理异常时才会被调用.
        只有在handle中, 你才能处理一些真正耗时的部分,
        比如复杂的ast解析等.
        大量的处理器在初始化时一个个做这件事是很不划算的.
        """
        yield from super().handle(ktb_exc)
        yield rc.translate("exthandler.pyyaml.hint") + "\n"


import pytest

import kawaiitb
from test.utils.utils import KTBTestBase, setup_test


class TestExceptionFormatting(KTBTestBase, console_output=True):
    def test_recursive_exception(self):
        """测试递归异常深度溢出"""
        setup_test()
        try:
            f = lambda x: f(x + 1)
            f(1)
        except Exception as e:
            tb = "".join(kawaiitb.traceback.format_exception(e))
            self.try_print_exc(e)
            assert "*在那之后" in tb

    def test_exception_with_notes(self):
        """测试带__notes__的异常"""
        class CustomException(Exception):
            def __init__(self, *args):
                super().__init__(*args)
                self.__notes__ = ["note1", "note2"]

        try:
            raise CustomException("test")
        except Exception as e:
            tb = "".join(kawaiitb.traceback.format_exception(e))
            self.try_print_exc(e)
            assert "test.test_tb_format.TestExceptionFormatting.test_exception_with_notes.<locals>.CustomException" in tb
            assert "note1\nnote2" in tb

    def test_in_package_error(self):
        """测试项目中模块的异常"""
        try:
            from test.utils.utils import raise_error
            raise_error()
        except Exception as e:
            import os
            tb = "".join(kawaiitb.traceback.format_exception(e))
            self.try_print_exc(e)
            # 应当截断了项目路径
            assert r"C://" not in tb and r"C:\\" not in tb  # Windows
            assert "文件 " + os.sep not in tb  # Linux

    def test_lib_module_error(self):
        """测试标准库模块异常"""
        import os
        try:
            import asyncio
            async def f():
                await asyncio.sleep(0)
                raise Exception("test")

            asyncio.run(f())
        except Exception as e:
            tb = "".join(kawaiitb.traceback.format_exception(e))
            self.try_print_exc(e)
            assert "asyncio 模块" in tb  # 显示模块名
            # 截断到lib或者site-packages
            assert "asyncio" + os.sep in tb
            assert os.sep + "asyncio" not in tb

    @pytest.mark.skip(reason="TODO")
    def test_long_line_formatting(self):
        """测试超长行异常格式化"""
        try:
            raise Exception(
                "This is a very long exception message that should be properly formatted and wrapped when displayed in the traceback output.")
        except Exception as e:
            tb = "".join(kawaiitb.traceback.format_exception(e))
            self.try_print_exc(e)
            assert len(tb.splitlines()) > 1  # 检查是否自动换行

    @pytest.mark.skip(reason="非项目帧简化 TODO")
    def test_site_package_error(self):
        """测试隔着一大堆模块的异常"""
        import asyncio
        async def task():
            await asyncio.sleep(0)
            raise Exception("test")

        asyncio.run(task())


import kawaiitb
from .utils.utils import KTBTestBase


class TestAnchor(KTBTestBase, console_output=False):
    # 1. 语法错误测试
    def test_syntax_error_anchor_invalid_lex(self):
        """测试无效词法语法错误"""
        try:
            exec("what can i say?")  # noqa
        except SyntaxError as e:
            tb = "".join(kawaiitb.traceback.format_exception(e))
            self.try_print_exc(e)
            assert """
    what can i say?
         ^^^""" in tb

    def test_syntax_error_anchor_incomplete_expr(self):
        """测试不完整表达式语法错误"""
        try:
            exec("1 + 2 *")  # noqa
        except SyntaxError as e:
            tb = "".join(kawaiitb.traceback.format_exception(e))
            self.try_print_exc(e)
            assert """
    1 + 2 *
           ^""" in tb

    def test_type_error_anchor_binop(self):
        """测试二元操作类型错误锚"""
        try:
            _ = "s" + 1  # noqa
        except TypeError as e:
            tb = "".join(kawaiitb.traceback.format_exception(e))
            self.try_print_exc(e)
            assert """
    _ = "s" + 1  # noqa
        ~~~~^~~""" in tb

    def test_type_error_anchor_subscript(self):
        """测试取键类型错误锚"""
        try:
            d = {"a": 1}["b"]  # noqa
        except KeyError as e:
            tb = "".join(kawaiitb.traceback.format_exception(e))
            self.try_print_exc(e)
            assert """
    d = {"a": 1}["b"]  # noqa
        ~~~~~~~~^^^^^""" in tb

    def test_zero_division_anchor(self):
        """测试除零错误锚"""
        try:
            _ = 1 / 0  # noqa
        except ZeroDivisionError as e:
            tb = "".join(kawaiitb.traceback.format_exception(e))
            self.try_print_exc(e)
            assert """
    _ = 1 / 0  # noqa
        ~~^~~""" in tb

    def test_attribute_error_anchor(self):
        """测试属性错误锚"""
        class Person:
            def __init__(self, name, age):
                self.name = name
                self.age = age

        try:
            person = Person("Alice", 30)
            _ = person.nonexistent_attribute  # noqa
        except AttributeError as e:
            tb = "".join(kawaiitb.traceback.format_exception(e))
            self.try_print_exc(e)
            assert """
    _ = person.nonexistent_attribute  # noqa
        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^""" in tb




import sys
from typing import Generator

import astroid

from kawaiitb.kraceback import KTBException
from kawaiitb.kwihandler import ErrorSuggestHandler
from kawaiitb.runtimeconfig import rc
from kawaiitb.utils import safe_string
from kawaiitb.utils.fromtraceback import compute_suggestion_error

__all__ = [
    "SyntaxErrorSuggestHandler",
    "ImportErrorSuggestHandler",
    "NameAttributeErrorSuggestHandler",
    # 这些属于原生增强，优先级都是1.1

]


@KTBException.register
class SyntaxErrorSuggestHandler(ErrorSuggestHandler, priority=1.1):
    """
    本处理器模仿原生的语法错误处理器，为语法错误添加额外的锚点指示
    """

    def __init__(self, exc_type, exc_value, exc_traceback, *, limit=None,
                 lookup_lines=True, capture_locals=False, compact=False,
                 max_group_width=15, max_group_depth=10, _seen=None):
        super().__init__(exc_type, exc_value, exc_traceback)
        if exc_type and issubclass(exc_type, SyntaxError):
            exc_value: SyntaxError
            self.filename = exc_value.filename
            lno = exc_value.lineno
            self.lineno = str(lno) if lno is not None else None
            end_lno = exc_value.end_lineno
            self.end_lineno = str(end_lno) if end_lno is not None else None
            self.text = exc_value.text
            self.offset = exc_value.offset
            self.end_offset = exc_value.end_offset
            self.msg = exc_value.msg

    @classmethod
    def translation_keys(cls):
        return {}  # 翻译键均由默认配置提供，不需要额外的翻译键

    def can_handle(self, ktb_exc) -> bool:
        return issubclass(ktb_exc.exc_type, SyntaxError)

    def handle(self, ktb_exc) -> Generator[str, None, None]:
        r"""
        (-) Traceback (most recent call last):
        (-)   File "C:\Users\BPuffer\Desktop\kawaii-traceback\main.py", line 139:8, in <module>
        (-)     exec("what can i say?")
        (1)   File "<string>", line 1
        (2)     what can i say?
        (3)          ^^^
        (4) SyntaxError: invalid syntax (<string>, line 1)
        """
        if self.lineno is not None:
            # part (1)
            yield rc.translate("frame.location.without_name",
                               file=self.filename or "<string>",  # repr转义
                               lineno=self.lineno, )

        text = self.text
        if text is not None:
            rtext = text.rstrip('\n')
            ltext = rtext.lstrip(' \n\f')
            spaces = len(rtext) - len(ltext)
            # part (2)
            yield rc.translate("frame.location.linetext",
                               line=ltext)

            if self.offset is not None:
                offset = self.offset
                end_offset = self.end_offset if self.end_offset not in {None, 0} else offset
                if offset == end_offset or end_offset == -1:
                    end_offset = offset + 1

                colno = offset - 1 - spaces
                end_colno = end_offset - 1 - spaces
                if colno >= 0:
                    # part (3)
                    # caretspace = ((c if c.isspace() else ' ') for c in ltext[:colno])
                    # yield '    {}{}'.format("".join(caretspace), ('^' * (end_colno - colno) + "\n"))
                    anchor_len = end_colno - colno
                    yield rc.anchors('    ' + ' ' * colno, 0, 0, anchor_len, anchor_len, crlf=True)

        msg = self.msg or "<no detail available>"
        # part (4)
        yield from super().handle(ktb_exc)


@KTBException.register
class ImportErrorSuggestHandler(ErrorSuggestHandler, priority=1.1):
    """
    本处理器模仿原生的ImportError的拼写错误检测
    为导入中的拼写错误添加额外的正确拼写提示
    """

    def __init__(self, exc_type, exc_value, exc_traceback, *, limit=None,
                 lookup_lines=True, capture_locals=False, compact=False,
                 max_group_width=15, max_group_depth=10, _seen=None):
        super().__init__(exc_type, exc_value, exc_traceback)

        self._can_handle = issubclass(exc_type, ImportError) and getattr(exc_value, "name_from", None) is not None

        self.suggestion = None
        if self._can_handle:
            self.wrong_name = getattr(exc_value, "name_from")
            self.suggestion = compute_suggestion_error(exc_value, exc_traceback, self.wrong_name)

    def can_handle(self, ktb_exc) -> bool:
        return self._can_handle

    @classmethod
    def translation_keys(cls):
        return {
            "default": {
                "native.import_error_suggestion.hint": "Did you mean '{suggestion}'?",
            },
            "zh_hans": {
                "native.import_error_suggestion.hint": "你可能是想导入'{suggestion}'",
            }
        }

    def handle(self, ktb_exc) -> Generator[str, None, None]:
        yield from super().handle(ktb_exc)
        if self.suggestion:
            yield rc.translate("native.import_error_suggestion.hint", suggestion=self.suggestion)


# @KTBException.register  # 正在考虑完全移除，或使用更通用的处理器 NameError和AttributeError 的Handler替代
class NameAttributeErrorSuggestHandler(ErrorSuggestHandler, priority=1.1):
    """
    本处理器模仿原生的NameError的拼写错误检测
    为NameError的拼写错误添加额外的正确拼写提示
    并为存在于标准库和第三方库中的名字添加额外的提示
    """

    def __init__(self, exc_type, exc_value, exc_traceback, *, limit=None,
                 lookup_lines=True, capture_locals=False, compact=False,
                 max_group_width=15, max_group_depth=10, _seen=None):
        super().__init__(exc_type, exc_value, exc_traceback)

        self._can_handle = (issubclass(exc_type, (NameError, AttributeError)) and
                            getattr(exc_value, "name", None) is not None)

        if self._can_handle:
            self.wrong_name = getattr(exc_value, "name")
            self.suggestion = compute_suggestion_error(exc_value, exc_traceback, self.wrong_name)
            self.is_stdlib = self.wrong_name in sys.stdlib_module_names

            self.is_3rd_party = False
            import importlib.metadata
            try:
                importlib.metadata.distribution(self.wrong_name)
                self.is_3rd_party = True
            except importlib.metadata.PackageNotFoundError:
                pass

            self.is_lib = self.is_stdlib or self.is_3rd_party

    def can_handle(self, ktb_exc) -> bool:
        return self._can_handle

    @classmethod
    def translation_keys(cls):
        return {
            "default": {
                "native.nameattr_error_suggestion.typo": "Did you mean '{suggestion}'?",
                "native.nameattr_error_suggestion.forget_import": "You may forget to import '{wrong_name}'",
                "native.nameattr_error_suggestion.or_forget_import": "or you may forget to import '{wrong_name}'",
            },
            "zh_hans": {
                "native.nameattr_error_suggestion.typo": "你是不是想输入'{suggestion}'？",
                "native.nameattr_error_suggestion.forget_import": "你可能忘记导入'{wrong_name}'了",
                "native.nameattr_error_suggestion.or_forget_import": "或者你可能忘记导入'{wrong_name}'了",
            }
        }

    def handle(self, ktb_exc) -> Generator[str, None, None]:
        yield from super().handle(ktb_exc)
        if self.suggestion:
            yield rc.translate("native.nameattr_error_suggestion.typo", suggestion=self.suggestion)

        if issubclass(ktb_exc.exc_type, NameError) and self.is_stdlib:
            if self.suggestion:
                yield rc.translate("native.nameattr_error_suggestion.or_forget_import", wrong_name=self.wrong_name)
            else:
                yield rc.translate("native.nameattr_error_suggestion.forget_import", wrong_name=self.wrong_name)

# 以上是所有原生处理中含有新增的处理逻辑的处理器。优先级均为1.1。


@KTBException.register
class StopIterationHandler(ErrorSuggestHandler, priority=1.0):  # 原生
    """
    StopIteration异常处理器
    ```
>>> def f():
>>>     for i in range(10):
>>>         yield i
>>>     return "Boom!"
>>>
>>> g = f()
>>> while True:
>>>     next(g)

... Traceback (most recent call last):
...   File "main.py", line 139, in <module>
...     next(g)
(-) StopIteration: Boom!

    改为:

(1) [StopIteration] 生成器'g'停止迭代: Boom!
    ```
    """

    def __init__(self, exc_type, exc_value, exc_traceback, **kwargs):
        super().__init__(exc_type, exc_value, exc_traceback, **kwargs)
        self._can_handle = issubclass(exc_type, StopIteration)
        if not self._can_handle:
            return

        # Python 3.7 之后，对于使用return 关键字的生成器，抛出的StopIteration异常会包含 return 的值。
        self.return_value = exc_value.value if hasattr(exc_value, 'value') else None


    def can_handle(self, ktb_exc) -> bool:
        return self._can_handle

    @classmethod
    def translation_keys(cls):
        return {
            "default": {
                "native.StopIteration.hint": "Generator '{generator}' stopped iterating.",
                "native.StopIteration.hint_with_return": "Generator '{generator}' stopped: {ret}",
            },
            "zh_hans": {
                "native.StopIteration.hint": "生成器'{generator}'没有更多值了。",
                "native.StopIteration.hint_with_return": "生成器'{generator}'没有更多值了: {ret}",
            }
        }

    def handle(self, ktb_exc: KTBException) -> Generator[str, None, None]:
        # 从栈帧中获取生成器在代码里的名称
        self.generator = "<...>"
        if len(ktb_exc.stack) > 0:
            exc_frame = ktb_exc.stack[0]
            for node in self.parse_ast_from_exc(exc_frame):
                # case: next(g) -> Call(
                #     func=Name(name=next),
                #     args=[...])
                if (
                        isinstance(node, astroid.Call) and  # 是函数调用
                        isinstance(node.func, astroid.Name) and  # 是显式函数名
                        node.func.name == 'next'  # 是next调用
                ):
                    self.generator = node.args[0].as_string()
                    break

                # case: g.__next__() -> Call(
                #     func=Attr(
                #         expr=<?>,
                #         attrname=__next__),
                #     args=[...])
                if (
                    isinstance(node, astroid.Call) and  # 是函数调用
                    isinstance(node.func, astroid.Attribute) and  # 是属性访问
                    node.func.attrname == '__next__'  # 是__next__方法调用
                ):
                    # 获取生成器表达式字符串
                    self.generator = node.func.expr.as_string()
                    break

        if self.return_value is not None:
            hint = rc.translate("native.StopIteration.hint_with_return", generator=self.generator, ret=self.return_value)
        else:
            hint = rc.translate("native.StopIteration.hint", generator=self.generator)
        yield rc.exc_line("StopIteration", hint)


@KTBException.register
class StopAsyncIterationHandler(ErrorSuggestHandler, priority=1.0):  # 原生
    """
    StopAsyncIteration异常处理器
    ```
>>> async def f():
>>>     for i in range(10):
>>>         yield i
>>>     return "Boom!"
>>>
>>> async for i in f():
>>>     pass

... Traceback (most recent call last):
...   File "main.py", line 139, in <module>
...     async for i in f():
(-) StopAsyncIteration: Boom!

    改为:

(1) [StopAsyncIteration] 异步生成器'f'停止迭代: Boom!
    ```
    """

    def __init__(self, exc_type, exc_value, exc_traceback, **kwargs):
        super().__init__(exc_type, exc_value, exc_traceback, **kwargs)
        self._can_handle = issubclass(exc_type, StopAsyncIteration)
        if not self._can_handle:
            return

        self.return_value = exc_value.value if hasattr(exc_value, 'value') else None

    def can_handle(self, ktb_exc) -> bool:
        return self._can_handle

    @classmethod
    def translation_keys(cls):
        return {
            "default": {
                "native.StopAsyncIteration.hint": "Async generator '{generator}' stopped iterating.",
                "native.StopAsyncIteration.hint_with_return": "Async generator '{generator}' stopped: {ret}",
            },
            "zh_hans": {
                "native.StopAsyncIteration.hint": "异步生成器'{generator}'没有更多值了。",
                "native.StopAsyncIteration.hint_with_return": "异步生成器'{generator}'没有更多值了: {ret}",
            }
        }

    def handle(self, ktb_exc) -> Generator[str, None, None]:
        # 从栈帧中获取异步生成器在代码里的名称
        self.generator = "<...>"
        if len(ktb_exc.stack) > 0:
            exc_frame = ktb_exc.stack[0]
            for node in self.parse_ast_from_exc(exc_frame):
                # case: anext(g) -> Call(
                #     func=Name(name=anext),
                #     args=[...])
                if (
                    isinstance(node, astroid.Call) and  # 是函数调用
                    isinstance(node.func, astroid.Name) and  # 是显式函数名
                    node.func.name == 'anext'  # 是anext调用
                ):
                    self.generator = node.args[0].as_string()
                    break

                # case: g.__anext__() -> Call(
                #     func=Attr(
                #         expr=<?>,
                #         attrname=__anext__),
                #     args=[...])
                if (
                    isinstance(node, astroid.Call) and  # 是函数调用
                    isinstance(node.func, astroid.Attribute) and  # 是属性访问
                    node.func.attrname == '__anext__'  # 是__anext__方法调用
                ):
                    # 获取异步生成器表达式字符串
                    self.generator = node.func.expr.as_string()
                    break

                # case: async for i in g: -> AsyncFor(
                #     target=<?>,
                #     iter=<?>,
                #     body=[...])
                if isinstance(node, astroid.AsyncFor):
                    self.generator = node.iter.as_string()
                    break

        if self.return_value is not None:
            hint = rc.translate("native.StopAsyncIteration.hint_with_return",
                             generator=self.generator,
                             ret=self.return_value)
        else:
            hint = rc.translate("native.StopAsyncIteration.hint",
                             generator=self.generator)
        if self.generator:
            yield rc.exc_line("StopAsyncIteration", hint)
        else:
            yield rc.exc_line("StopAsyncIteration", hint)


@KTBException.register
class OverflowErrorHandler(ErrorSuggestHandler, priority=1.0):
    """
    OverflowError异常处理器
    ```
>>> import math
>>> math.exp(1000)

... Traceback (most recent call last):
...   File "<input>", line 1, in <module>
(-) OverflowError: math range error

    改为:

(1) [OverflowError] 溢出错误: 数学范围错误
    ```
    """

    def __init__(self, exc_type, exc_value, exc_traceback, **kwargs):
        super().__init__(exc_type, exc_value, exc_traceback, **kwargs)
        self._can_handle = issubclass(exc_type, OverflowError)
        if self._can_handle:
            self.err_msg_key = {
                "math range error": "native.OverflowError.msg.math_range_error",
            }.get(safe_string(exc_value, '<exception>'), exc_value or "native.OverflowError.msg.novalue")  # match None and ""


    def can_handle(self, ktb_exc) -> bool:
        return self._can_handle

    @classmethod
    def translation_keys(cls):
        return {
            "default": {
                "native.OverflowError.msg.novalue": "A value is too large for the given type",
                "native.OverflowError.msg.math_range_error": "math range error",
            },
            "zh_hans": {
                "native.OverflowError.msg.novalue": "数值超出了其类型所能表示的范围",
                "native.OverflowError.msg.math_range_error": "数学范围错误",
            }
        }

    def handle(self, ktb_exc) -> Generator[str, None, None]:
        self.err_msg = rc.translate(self.err_msg_key)
        yield rc.exc_line("OverflowError", self.err_msg)


@KTBException.register
class ZeroDivisionErrorHandler(ErrorSuggestHandler, priority=1.0):
    """
    ZeroDivisionError异常处理器
    ```
>>> 1 / (1 - 1)

... Traceback (most recent call last):
...   File "<input>", line 1, in <module>
(-) ZeroDivisionError: division by zero

    改为:

(1) [ZeroDivisionError] 除以零 - '(1 - 1)'的值为0
    ```
    """
    def __init__(self, exc_type, exc_value, exc_traceback, **kwargs):
        super().__init__(exc_type, exc_value, exc_traceback, **kwargs)
        self._can_handle = issubclass(exc_type, ZeroDivisionError)
        self.exc_value = exc_value
        self.stack = exc_traceback

    def can_handle(self, ktb_exc) -> bool:
        return self._can_handle

    @classmethod
    def translation_keys(cls):
        return {
            "default": {
                "native.ZeroDivisionError.msg_plain": "division by zero",
                "native.ZeroDivisionError.msg": "division by zero - '{divisor}' evaluates to 0",
                "native.ZeroDivisionError.easter_eggs":["KawaiiTraceback has been installed successfully!",
                                                         "Congratulations! You have successfully installed KawaiiTraceback!",
                                                         "You can't divide by zero! QwQ",
                                                         "Tips: TracebackException is the only Exception that cannot be raised.",
                                                         "1/0 = ∞ (in the Riemann sphere)"]
            },
            "zh_hans": {
                "native.ZeroDivisionError.msg_plain": "除以零",
                "native.ZeroDivisionError.msg": "除以零 - '{divisor}'的值为0",
                "native.ZeroDivisionError.easter_eggs": ["KawaiiTraceback安装成功!",
                                                         "恭喜你发现了Python的隐藏特性：无限能量生成器！",
                                                         "不可以除以零啦喵~(´•ω•̥`)",
                                                         "冷知识：TracebackException是唯一一个不能raise的Exception",
                                                         "1/0 = ∞ (在黎曼球面上成立)"]
            }
        }

    def handle(self, ktb_exc) -> Generator[str, None, None]:
        # 一般ZeroDivisionError的错误信息都是division by zero, 或者没有。这两种情况都可以直接用翻译
        for node in self.parse_ast_from_exc(ktb_exc.stack[0]):
            # case: 1 / 0 -> BinOp(
            #     left=Num(n=1),
            #     op=Div(),
            #     right=Num(n=0))
            if (
                isinstance(node, astroid.BinOp) and node.op == '/' and  # 是转浮点除
                isinstance(node.left, astroid.Const) and node.left.value == 1 and  # 被除数是1
                isinstance(node.right, astroid.Const) and node.right.value == 0  # 除数是0
            ):
                # 输入1/0触发彩蛋
                import random
                egg = random.choice(rc.translate("native.ZeroDivisionError.easter_eggs"))
                yield rc.exc_line("KawaiiTB", egg)
                break
            elif(
                isinstance(node, astroid.BinOp) and node.op in ('/', '//')  # 是除法
            ):
                # 这就够了。不需要太麻烦的匹配，二元操作的错误帧定位本身就很精准了
                if self.exc_value is None or \
                        safe_string(self.exc_value, "") == "division by zero" or \
                        safe_string(self.exc_value, "") == "float division by zero":
                    hint = rc.translate("native.ZeroDivisionError.msg", divisor=node.right.as_string())
                else:
                    hint = self.exc_value
                yield rc.exc_line("ZeroDivisionError", hint)
                break
        else:
            yield rc.exc_line("ZeroDivisionError", rc.translate("native.ZeroDivisionError.msg_plain"))


@KTBException.register
class AssertionErrorHandler(ErrorSuggestHandler, priority=1.0):
    """
    AssertionError异常处理器
    ```
>>> a, b = 1, 2
>>> assert a == b

... Traceback (most recent call last):
...   File "<input>", line 1, in <module>
(-) AssertionError
    改为:
(1) [AssertionError] 断言 a == b, 但是 a=1, b=2.
    ```
    """
    def __init__(self, exc_type, exc_value, exc_traceback, **kwargs):
        super().__init__(exc_type, exc_value, exc_traceback, **kwargs)
        self._can_handle = issubclass(exc_type, AssertionError)
        self.exc_value = exc_value
        self.exc_traceback = exc_traceback

    def can_handle(self, ktb_exc) -> bool:
        return self._can_handle

    @classmethod
    def translation_keys(cls):
        return {
            "default": {
                "native.AssertionError.msg": "Assertion {assertion} failed.",
                "native.AssertionError.msg_with_values": "Assertion {assertion}, but {values}."
            },
            "zh_hans": {
                "native.AssertionError.msg": "断言 {assertion} 失败。",
                "native.AssertionError.msg_with_values": "断言 {assertion}, 但是 {values}."
            }
        }

    def handle(self, ktb_exc: KTBException) -> Generator[str, None, None]:
        # 如果有信息，直接返回信息
        if (self.exc_value is not None and safe_string(self.exc_value, "") != ""
            or len(ktb_exc.stack) == 0):
            yield rc.exc_line("AssertionError", safe_string(self.exc_value, "<exception>"))
            return

        # 从栈帧中获取断言的表达式字符串
        assert_expr = None
        assert_exprs: set[str] = set()
        exc_frame = ktb_exc.stack[0]
        for node in self.parse_ast_from_exc(exc_frame, parse_line=True):
            if not isinstance(node, astroid.Assert):
                continue
            expr = node.test
            assert_expr = expr.as_string()
            if not assert_expr:
                continue
            # 收集断言表达式中的变量名
            if isinstance(expr, astroid.Compare):
                # 比较表达式: a == b, a > b > c 等
                assert_exprs.add(expr.left.as_string())
                for _, right in expr.ops:
                    assert_exprs.add(right.as_string())
                    # TODO: 支持递归的表达式
                    # 问题: 如何判断递归下的表达式是用户所需要看到的
                    # 阻力: 断言表达式通常大道至简, 甚至第二层嵌套都很少看到, 实用型存疑
            elif isinstance(expr, astroid.BoolOp):
                # 布尔运算: a and b and c 等
                assert_exprs.add(expr.as_string())
                for value in expr.values:
                    if isinstance(value, astroid.Name):
                        assert_exprs.add(value.as_string())
            elif isinstance(expr, astroid.Name):
                # 简单变量: assert a
                assert_exprs.add(expr.as_string())
            elif isinstance(expr, astroid.Call):
                # 函数调用: assert a()
                # 如果函数以"is""not""has"开头, 则取得所有函数参数, 否则只取整个表达式的值
                if isinstance(expr.func, astroid.Name):
                    func_name = expr.func.as_string().split(".")[-1]
                    if func_name.startswith(("is", "not", "has")):
                        [
                            assert_exprs.add(arg.as_string())
                            for arg in expr.args
                            if isinstance(arg, (astroid.Name, astroid.Expr))
                        ]
                        # for arg in expr.args:
                        #     assert_exprs.add(arg.as_string())
                    else:
                        assert_exprs.add(expr.as_string())
            elif isinstance(expr, (astroid.BinOp, astroid.UnaryOp)):
                # 一二元运算等直接求值, 这些变量的最终值不是布尔, 用户只需要这个值.
                for operand in [expr.left, expr.right]:
                    if isinstance(operand, astroid.Name):
                        assert_exprs.add(operand.as_string())
            break

        if assert_expr is None:
            yield rc.exc_line("AssertionError", rc.translate("native.AssertionError.msg", assertion="<Unknown Expression>"))
            return

        # 获取变量的实际值
        values = []
        frame = self.exc_traceback.tb_frame
        globals_dict = frame.f_globals
        locals_dict = frame.f_locals

        for expr_str in assert_exprs:
            try:
                evaluated = eval(expr_str, globals_dict, locals_dict)
                # 此处使用eval是因为原表达式一定已经求值成功了，才会报AssertionError
                values.append(f"{expr_str}={evaluated!r}")
            except Exception:
                # 理论上不会进入这个分支, 但安全起见.
                yield f"[KawaiiTB Error] strange error when eval {expr_str}"

        values_str = ", ".join(values)
        if len(values_str) > 50:
            values_str = values_str[:50] + "..."

        if values:
            yield rc.exc_line(
                "AssertionError",
                rc.translate("native.AssertionError.msg_with_values",
                             assertion=assert_expr,
                             values=", ".join(values)))
        else:
            yield rc.exc_line(
                "AssertionError",
                rc.translate(
                    "native.AssertionError.msg",
                    assertion=assert_expr
                )
            )


@KTBException.register
class EOFErrorHandler(ErrorSuggestHandler, priority=1.0):
    r"""
    EOFError异常处理器
    ```
>>> import sys
>>> from io import StringIO
>>> sys.stdin = StringIO("")
>>> input()

... Traceback (most recent call last):
...   File "<input>", line 1, in <module>
(-) EOFError
    改为:
(1) [EOFError] 输入结束
    ```
    """
    def __init__(self, exc_type, exc_value, exc_traceback, **kwargs):
        super().__init__(exc_type, exc_value, exc_traceback, **kwargs)
        self._can_handle = issubclass(exc_type, EOFError)
        if self._can_handle:
            self.err_msg_key = {
                "EOF when reading a line": "native.EOFError.when_reading_line",
            }.get(safe_string(exc_value, '<exception>'), exc_value or "native.EOFError.msg")

    def can_handle(self, ktb_exc) -> bool:
        return self._can_handle

    @classmethod
    def translation_keys(cls):
        return {
            "default": {
                "native.EOFError.when_reading_line": "EOF when reading a line",
                "native.EOFError.msg": "get stop signal, data reading ended",
            },
            "zh_hans": {
                "native.EOFError.when_reading_line": "读取一行时接收到了停止信号",
                "native.EOFError.msg": "接收到了停止信号，数据读取结束",
            }
        }

    def handle(self, ktb_exc) -> Generator[str, None, None]:
        self.err_msg = rc.translate(self.err_msg_key)
        yield rc.exc_line("EOFError", self.err_msg)


@KTBException.register
class KeyboardInterruptHandler(ErrorSuggestHandler, priority=1.0):
    """
    KeyboardInterrupt异常处理器

    当用户按下Ctrl+C时，会抛出KeyboardInterrupt异常
    ```
>>> # 用户按下Ctrl+C

...
... Traceback (most recent call last):
...   File "<input>", line 1, in <module>
(-) KeyboardInterrupt
    改为:
(1) [KeyboardInterrupt] 手动终止了程序
    ```
    """

    def __init__(self, exc_type, exc_value, exc_traceback, **kwargs):
        super().__init__(exc_type, exc_value, exc_traceback, **kwargs)
        self._can_handle = issubclass(exc_type, KeyboardInterrupt)

    def can_handle(self, ktb_exc) -> bool:
        return self._can_handle

    @classmethod
    def translation_keys(cls):
        return {
            "default": {
                "native.KeyboardInterrupt.msg": "Program interrupted by user",
            },
            "zh_hans": {
                "native.KeyboardInterrupt.msg": "手动终止了程序",
                "native.KeyboardInterrupt.msg_extra": "手动终止了程序：{extra}",
            }
        }

    def handle(self, ktb_exc) -> Generator[str, None, None]:
        if not str(ktb_exc).strip():
            yield rc.exc_line("KeyboardInterrupt", rc.translate("native.KeyboardInterrupt.msg"))
        else:
            yield rc.exc_line("KeyboardInterrupt", rc.translate("native.KeyboardInterrupt.msg_extra", extra=str(ktb_exc)))

class SystemExitHandler(ErrorSuggestHandler, priority=1.0):
    """
    SystemExit异常处理器
>>> exit(114514)

... Traceback (most recent call last):
...   File "<input>", line 1, in <module>
(-) SystemExit: 114514
    改为:
(1) [SystemExit] 程序退出
    """
    def __init__(self, exc_type, exc_value, exc_traceback, **kwargs):
        super().__init__(exc_type, exc_value, exc_traceback, **kwargs)
        self._can_handle = issubclass(exc_type, SystemExit)

TODOS = {
    "BaseException": {
        "BaseException": "不设计。这个异常过于抽象，基本没人会单独抛",
        "SystemExit": "不设计。这个异常不应该是被捕获的。",
        "KeyboardInterrupt": "KeyboardInterruptHandler(ErrorSuggestHandler, priority=1.0)",  # Complete
        "GeneratorExit": "不要设计，因为这个异常不会被显示",
        "Exception": {
            "Exception": "不设计。这个异常过于抽象，基本没人会单独抛",
            "StopIteration": "StopIterationHandler(ErrorSuggestHandler, priority=1.0)",  # Complete
            "StopAsyncIteration": "StopAsyncIterationHandler(ErrorSuggestHandler, priority=1.0)",  # Complete
            "ArithmeticError": {
                "ArithmeticError": "不设计。这个异常过于抽象，基本没人会单独抛",
                "FloatingPointError": "不设计。这个异常应当不再出现。",
                "OverflowError": "OverflowErrorHandler(ErrorSuggestHandler, priority=1.0)",  # Complete
                "ZeroDivisionError": "ZeroDivisionErrorHandler(ErrorSuggestHandler, priority=1.0)",  # Complete
            },
            "AssertionError": "AssertionErrorHandler(ErrorSuggestHandler, priority=1.0)",  # Complete
            "AttributeError": "AttributeErrorHandler(ErrorSuggestHandler, priority=1.0)",  # Complete
            "BufferError": "过于过于罕见了，能碰见的基本都是在玩底层的人，没必要给他们讲解，不设计",
            "EOFError": "EOFErrorHandler(ErrorSuggestHandler, priority=1.0)",  # Complete
            "ImportError": {
                "ImportError": "ImportErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                "ModuleNotFoundError": "ModuleNotFoundErrorHandler(ErrorSuggestHandler, priority=1.05)",  # TODO
            },
            "LookupError": {
                "LookupError": "不设计。这个异常过于抽象，基本没人会单独抛",
                "IndexError": "IndexErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                "KeyError": "KeyErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
            },
            "MemoryError": "MemoryErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
            "NameError": {
                "NameError": "NameErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                "UnboundLocalError": "UnboundLocalErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
            },
            "OSError": {
                "OSError": "OSErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                # 这个类别只设计几个常见的，太少见的就不设计了
                "BlockingIOError": "",
                "ChildProcessError": "",
                "ConnectionError": {
                    "ConnectionError": "",
                    "BrokenPipeError": "",
                    "ConnectionAbortedError": "",
                    "ConnectionRefusedError": "",
                    "ConnectionResetError": "",
                },
                "FileExistsError": "FileExistsErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                "FileNotFoundError": "FileNotFoundErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                "InterruptedError": "",
                "IsADirectoryError": "IsADirectoryErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                "NotADirectoryError": "NotADirectoryErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                "PermissionError": "PermissionErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                "ProcessLookupError": "",
                "TimeoutError": "",
            },
            "ReferenceError": "实在过于罕见，疑似cpy完备化接口产物，不设计",
            "RuntimeError": {
                "RuntimeError": "RuntimeErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                "NotImplementedError": "NotImplementedErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                "RecursionError": "RecursionErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
            },
            "SyntaxError": {
                "SyntaxError": "SyntaxErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                "IndentationError": {
                    "IndentationError": "IndentationErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                    "TabError": "TabErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                },
            },
            "SystemError": "SystemErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
            "TypeError": "TypeErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
            "ValueError": {
                "ValueError": "ValueErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                "UnicodeError": {
                    "UnicodeError": "UnicodeErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                    "UnicodeDecodeError": "UnicodeDecodeErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                    "UnicodeEncodeError": "UnicodeEncodeErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                    "UnicodeTranslateError": "UnicodeTranslateErrorHandler(ErrorSuggestHandler, priority=1.0)",  # TODO
                },
            }
        }
    }
}


from .defaults import *
from .attribute_handler import AttributeErrorHandler
from .extensions import *

# __all__ = [
#     "AttributeErrorHandler",
#     ...  # TODO: 显式列出所有的处理器. 等全写完再来写这个.
# ]
__all__ = [
    "AttributeErrorHandler",
]


"""
Kawaii Traceback - Cute Python traceback beautifier

A delightful tool for transforming standard Python tracebacks into adorable and 
user-friendly error messages with multilingual support.
"""

__version__ = "0.1.0"
__author__ = "BPuffer <mc-puffer@qq.com>"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2025 BPuffer"
__description__ = "A kawaii Python traceback beautifier with multilingual support"

import sys

import kawaiitb.kraceback as traceback
from kawaiitb.handlers import __all__ as _handlers
from kawaiitb.kraceback import KTBException
from kawaiitb.kwihandler import ErrorSuggestHandler
from kawaiitb.runtimeconfig import rc, load_config
from kawaiitb.tools import load, unload

__all__ = [
    "traceback",
    "rc",
    "load",
    "unload",
    "ErrorSuggestHandler",
    "KTBException",
    "load_config",
    *_handlers,
]


"""
本文件重写了traceback中的部分结构，使其支持自定义的异常处理器。
结构与traceback.py相比有减改而无增。

"Extract, format and print information about Python stack traces."
"""

import collections.abc
import itertools
import linecache
import os
import sys
import sysconfig
from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Type, TYPE_CHECKING, Generator

from kawaiitb.runtimeconfig import rc
from kawaiitb.utils import sys_getframe, extract_caret_anchors_from_line_segment, safe_string
from kawaiitb.utils.fromtraceback import (
    sentinel, parse_value_tb, walk_tb_with_full_positions,
    byte_offset_to_character_offset, walk_stack,
    ExceptionPrintContext, display_width,
)

if TYPE_CHECKING:
    from kawaiitb.kwihandler import ErrorSuggestHandler


class _ENV:
    def __init__(self):
        self.update()

    def update(self):
        """更新环境信息"""
        self.cwd = os.getcwd()
        self.platform = sys.platform

    def __getattr__(self, name):
        """按需获取新属性时自动更新"""
        self.update()
        return super().__getattribute__(name)

ENV = _ENV()


#
# 格式化和打印回溯信息
#

def print_list(extracted_list, file=None):
    """Print the list of tuples as returned by extract_tb() or
    extract_stack() as a formatted stack trace to the given file."""
    if file is None:
        file = sys.stderr
    for item in StackSummary.from_list(extracted_list).format():
        print(item, file=file, end="")


def format_list(extracted_list):
    """Format a list of tuples or FrameSummary objects for printing.

    Given a list of tuples or FrameSummary objects as returned by
    extract_tb() or extract_stack(), return a list of strings ready
    for printing.

    Each string in the resulting list corresponds to the item with the
    same index in the argument list.  Each string ends in a newline;
    the strings may contain internal newlines as well, for those items
    whose source text line is not None.
    """
    return StackSummary.from_list(extracted_list).format()


#
# 打印和提取异常
#

def print_tb(tb, limit=None, file=None):
    """Print up to 'limit' stack trace entries from the traceback 'tb'.

    If 'limit' is omitted or None, all entries are printed.  If 'file'
    is omitted or None, the output goes to sys.stderr; otherwise
    'file' should be an open file or file-like object with a write()
    method.
    """
    print_list(extract_tb(tb, limit=limit), file=file)


def format_tb(tb, limit=None):
    """A shorthand for 'format_list(extract_tb(tb, limit))'."""
    return extract_tb(tb, limit=limit).format()


def extract_tb(tb, limit=None):
    """
    Return a StackSummary object representing a list of
    pre-processed entries from traceback.

    This is useful for alternate formatting of stack traces.  If
    'limit' is omitted or None, all entries are extracted.  A
    pre-processed stack trace entry is a FrameSummary object
    containing attributes filename, lineno, name, and line
    representing the information that is usually printed for a stack
    trace.  The line is a string with leading and trailing
    whitespace stripped; if the source is not available it is None.
    """
    return StackSummary.extract_from_extended_frame_gen(
        walk_tb_with_full_positions(tb), limit=limit)


#
# 异常格式化与输出
#


def print_exception(exc, /, value=sentinel, tb=sentinel, limit=None, file=None, chain=True):
    """Print exception up to 'limit' stack trace entries from 'tb' to 'file'.

    This differs from print_tb() in the following ways: (1) if
    traceback is not None, it prints a header "Traceback (most recent
    call last):"; (2) it prints the exception type and value after the
    stack trace; (3) if type is SyntaxError and value has the
    appropriate format, it prints the line where the syntax error
    occurred with a caret on the next line indicating the approximate
    position of the error.
    """
    value, tb = parse_value_tb(exc, value, tb)
    te = KTBException(type(value), value, tb, limit=limit, compact=True)
    te.print(file=file, chain=chain)


def format_exception(exc, /, value=sentinel, tb=sentinel, limit=None, chain=True):
    """Format a stack trace and the exception information.

    The arguments have the same meaning as the corresponding arguments
    to print_exception().  The return value is a list of strings, each
    ending in a newline and some containing internal newlines.  When
    these lines are concatenated and printed, exactly the same text is
    printed as does print_exception().
    """
    value, tb = parse_value_tb(exc, value, tb)
    te = KTBException(type(value), value, tb, limit=limit, compact=True)
    return list(te.format(chain=chain))


def format_exception_only(exc, /, value=sentinel):
    """Format the exception part of a traceback.

    The return value is a list of strings, each ending in a newline.

    The list contains the exception's message, which is
    normally a single string; however, for :exc:`SyntaxError` exceptions, it
    contains several lines that (when printed) display detailed information
    about where the syntax error occurred. Following the message, the list
    contains the exception's ``__notes__``.
    """
    if value is sentinel:
        value = exc
    te = KTBException(type(value), value, None, compact=True)
    return list(te.format_exception_only())


# --

def print_exc(limit=None, file=None, chain=True):
    """Shorthand for 'print_exception(sys.exception(), limit, file, chain)'."""
    print_exception(sys.exception(), limit=limit, file=file, chain=chain)


def format_exc(limit=None, chain=True):
    """Like print_exc() but return a string."""
    return "".join(format_exception(sys.exception(), limit=limit, chain=chain))


def print_last(limit=None, file=None, chain=True):
    """This is a shorthand for 'print_exception(sys.last_exc, limit, file, chain)'."""
    if not hasattr(sys, "last_exc") and not hasattr(sys, "last_type"):
        raise ValueError("no last exception")

    if hasattr(sys, "last_exc"):
        print_exception(sys.last_exc, limit, file, chain)
    else:
        print_exception(sys.last_type, sys.last_value, sys.last_traceback,
                        limit, file, chain)


#
# 堆栈打印与提取
#

def print_stack(f=None, limit=None, file=None):
    """从调用点打印堆栈跟踪信息。

    可选参数'f'可指定起始堆栈帧，'limit'和'file'参数
    与print_exception()具有相同含义。
    """
    if f is None:
        f = sys_getframe().f_back
    print_list(extract_stack(f, limit=limit), file=file)


def format_stack(f=None, limit=None):
    """'format_list(extract_stack(f, limit))'的快捷方式"""
    if f is None:
        f = sys_getframe().f_back
    return format_list(extract_stack(f, limit=limit))


def extract_stack(f=None, limit=None):
    """从当前堆栈帧提取原始回溯信息。

    返回值格式与extract_tb()相同。可选参数'f'和'limit'
    与print_stack()含义相同。返回列表中的每个元素是四元组
    (文件名, 行号, 函数名, 代码文本)，按从最旧到最新
    堆栈帧的顺序排列。
    """
    if f is None:
        f = sys_getframe().f_back
    stack = StackSummary.extract(walk_stack(f), limit=limit)
    stack.reverse()
    return stack


def clear_frames(tb):
    """清除回溯中所有帧对局部变量的引用。"""
    while tb is not None:
        try:
            tb.tb_frame.clear()
        except RuntimeError:
            # 忽略帧仍在执行时抛出的异常
            pass
        tb = tb.tb_next


@dataclass
class FrameSummary:
    """表示回溯中单个帧的信息。

    - :attr:`filename` 该帧对应的文件名
    - :attr:`lineno` 捕获该帧时在文件中活跃的代码行号
    - :attr:`name` 捕获该帧时正在执行的函数或方法名称
    - :attr:`line` 来自linecache模块的该帧执行时的源代码文本
    - :attr:`locals_` 如果没有提供则为None，否则是变量名到其repr()表示的字典
    """

    __slots__ = ('filename', 'lineno', 'end_lineno', 'colno', 'end_colno',
                 'name', '_line', 'locals_')

    filename: str
    lineno: int
    name: str
    end_lineno: Optional[int]
    colno: Optional[int]
    end_colno: Optional[int]
    _line: Optional[str]
    locals_: Optional[dict]

    def __init__(self, filename, lineno, name, *, lookup_line=True,
                 locals_=None, line=None,
                 end_lineno=None, colno=None, end_colno=None):
        """构造FrameSummary对象。

        :param lookup_line: 如果为True，会立即通过linecache查找源代码行。
            否则推迟到首次需要时再查找
        :param locals_: 如果提供，会捕获帧的局部变量作为对象表示
        :param line: 如果提供，直接使用该行文本而不通过linecache查找
        """
        self.filename = filename
        self.lineno = lineno
        self.name = name
        self._line = line
        if lookup_line:  # 立即获取源代码行
            self.load_line()
        self.locals_ = {k: safe_string(v, 'local', func=repr)
                        for k, v in locals_.items()} if locals_ else None
        self.end_lineno = end_lineno
        self.colno = colno
        self.end_colno = end_colno

    def __eq__(self, other):
        if isinstance(other, FrameSummary):
            return (self.filename == other.filename and
                    self.lineno == other.lineno and
                    self.name == other.name and
                    self.locals_ == other.locals_)
        if isinstance(other, tuple):
            return (self.filename, self.lineno, self.name, self.line) == other
        return NotImplemented

    def __getitem__(self, pos):
        return (self.filename, self.lineno, self.name, self.line)[pos]

    def __iter__(self):
        return iter([self.filename, self.lineno, self.name, self.line])

    def __repr__(self):
        return "<FrameSummary file {filename}, line {lineno} in {name}>".format(
            filename=self.filename, lineno=self.lineno, name=self.name)

    def __len__(self):
        return 4

    @property
    def original_line(self):
        # 返回源代码中的原始行文本，不strip
        self.load_line()
        return self._line

    def load_line(self):
        if self._line is None:
            if self.lineno is None:
                return None
            self._line = linecache.getline(self.filename, self.lineno)
        return self._line.strip()

    @property
    def line(self):
        return self.load_line()

def parse_module_filename(filename: str) -> str:
    """处理模块文件名，返回格式化后的显示字符串"""
    # 优先处理标准库路径
    stdlib_path = Path(sysconfig.get_path("stdlib")).resolve()
    try:
        abs_path = Path(filename).resolve().relative_to(stdlib_path)
        parts = abs_path.parts
        if parts:
            return rc.translate(
                "config.file.parsed_filename",
                namespace=parts[0],
                filename=os.sep.join(parts)
            )
    except ValueError:
        pass

    # 工作区文件处理
    cwd = Path(ENV.cwd).resolve()
    try:
        rel_path = Path(filename).resolve().relative_to(cwd)
        if not rc.translate("config.file.include_cwd"):
            return str(rel_path)
        return filename
    except ValueError:
        pass

    # 第三方包路径检测
    normalized = os.path.normpath(filename).replace("\\", "/").lower()
    patterns = [
        ("site-packages", 1),  # 匹配层级
        ("dist-packages", 1),
        ("lib/python", 2),     # 跳过版本号目录
        ("lib64/python", 2),
        ("/python", 2),       # Windows 路径如 C:/Python312/Lib/
    ]

    for pattern, skip in patterns:
        index = normalized.find(f"/{pattern}/")
        if index != -1:
            parts = normalized[index+1:].split("/")[skip+1:]
            if parts:
                return rc.translate(
                    "config.file.parsed_filename",
                    namespace=parts[0],
                    filename=os.sep.join(parts)
                )

    return filename


# _RECURSIVE_CUTOFF = 3  # Also hardcoded in traceback.c.


class StackSummary(list[FrameSummary]):
    """FrameSummary对象的list, 表示异常帧的栈"""

    @classmethod
    def extract(cls, frame_gen, *, limit=None, lookup_lines=True,
                capture_locals=False):
        """从回溯或堆栈对象创建StackSummary。

        :param frame_gen: 生成(frame, lineno)元组的生成器，这些元组的摘要将被包含在堆栈中。
        :param limit: None表示包含所有帧，或指定要包含的帧数。
        :param lookup_lines: 如果为True，立即查找每个帧的源代码行，否则推迟到帧被渲染时再查找。
        :param capture_locals: 如果为True，将捕获每个帧的局部变量。作为对象表示到FrameSummary中。
        """

        def extended_frame_gen():
            for f, lineno in frame_gen:
                yield f, (lineno, None, None, None)

        return cls.extract_from_extended_frame_gen(
            extended_frame_gen(), limit=limit, lookup_lines=lookup_lines,
            capture_locals=capture_locals)

    @classmethod
    def extract_from_extended_frame_gen(cls, frame_gen, *, limit=None,
                                        lookup_lines=True, capture_locals=False):
        """从扩展的帧生成器提取堆栈信息

        参数:
        - frame_gen: 生成器，生成(frame, (lineno, end_lineno, colno, end_colno))格式的元组
        - limit: 限制提取的帧数量，None表示无限制
        - lookup_lines: 是否立即查找源代码行
        - capture_locals: 是否捕获局部变量

        返回:
        - StackSummary对象，包含提取的帧信息
        """

        # 处理最大深度限制
        if limit is None:
            # 从系统获取默认的traceback限制
            limit = getattr(sys, 'tracebacklimit', None)
            if limit is not None and limit < 0:
                # 如果限制为负数，则设为0
                limit = 0

        if limit is not None:
            if limit >= 0:
                # 正数限制：使用itertools.islice截取前limit个帧
                frame_gen = itertools.islice(frame_gen, limit)
            else:
                # 负数限制：使用collections.deque保留最后limit个帧
                frame_gen = collections.deque(frame_gen, maxlen=-limit)

        result = cls()  # 创建结果对象
        fnames = set()  # 用于存储所有文件名，避免重复加载

        # 遍历frame_gen，构建FrameSummary对象
        for f, (lineno, end_lineno, colno, end_colno) in frame_gen:
            co = f.f_code  # 获取代码对象

            filename = co.co_filename  # 获取文件名
            fnames.add(filename)  # 添加文件名到集合中

            name = co.co_name  # 获取函数名

            linecache.lazycache(filename, f.f_globals)  # 延迟加载文件内容到linecache

            # 是否捕获局部变量
            if capture_locals:
                f_locals = f.f_locals
            else:
                f_locals = None

            # 创建FrameSummary对象并添加到结果中
            result.append(FrameSummary(
                filename, lineno, name, lookup_line=False, locals_=f_locals,
                end_lineno=end_lineno, colno=colno, end_colno=end_colno))

        # 检查并更新linecache中的文件内容
        for filename in fnames:
            linecache.checkcache(filename)

        # 如果需要立即查找，现在触发查找
        if lookup_lines:
            for f in result:
                f.load_line()  # 加载源代码行, 放入了FrameSummary的_line

        return result

    @classmethod
    def from_list(cls, a_list):
        """
        从提供的FrameSummary对象列表或旧式元组列表创建StackSummary对象。
        """
        # 虽然进行isinstance(a_list, StackSummary)的快速检查很诱人，
        # 但idlelib.run.cleanup_traceback和其他类似代码可能会通过将任意帧
        # 转换为普通元组来破坏这一点，因此我们需要逐帧检查。
        result = StackSummary()
        for frame in a_list:
            if isinstance(frame, FrameSummary):
                result.append(frame)
            else:
                filename, lineno, name, line = frame
                result.append(FrameSummary(filename, lineno, name, line=line))
        return result

    def format_frame_summary(self, frame_summary: FrameSummary):
        """格式化单个FrameSummary的行。

        返回表示堆栈中单个帧的字符串。对于要打印在堆栈摘要中的每个帧，
        都会调用此方法。
        """
        # 文件名，据此判断是否为工作区，并判断是否包含工作区路径
        filename = parse_module_filename(frame_summary.filename)

        # 函数名，判断是否为模块级语句并翻译
        name = frame_summary.name
        if name == "<module>":
            name = rc.translate("config.module")
        elif name == "<string>":
            name = rc.translate("config.string")

        # 格式化帧信息行
        if frame_summary.colno is not None:
            row = [rc.translate('frame.location.with_column',
                                file=filename,
                                name=name,
                                lineno=frame_summary.lineno,
                                colno=frame_summary.colno)]
        else:
            row = [rc.translate('frame.location.without_column',
                                file=frame_summary.filename,
                                lineno=frame_summary.lineno,
                                name=frame_summary.name)]

        # 如果这个帧有可用的行文本
        if frame_summary.line:
            # 添加行文本到结果中
            stripped_line = frame_summary.line.strip()
            row.append(rc.translate('frame.location.linetext', line=stripped_line))

            # 如果有列位置信息，显示指示锚
            if (
                    frame_summary.colno is not None
                    and frame_summary.end_colno is not None
            ):
                # a = "s" + 1  语法上'"s" + 1'是错误产生的位置
                #     ~~~~^~~  而主字符"+"是错误产生的根本原因。

                # 处理问题帧的代码段基本内容
                line = frame_summary.original_line
                orig_line_len = len(line)
                frame_line_len = len(frame_summary.line.lstrip())
                stripped_characters = orig_line_len - frame_line_len
                start_offset = byte_offset_to_character_offset(line, frame_summary.colno)
                end_offset = byte_offset_to_character_offset(line, frame_summary.end_colno)
                code_segment = line[start_offset:end_offset]

                # 如果是单行问题帧, 语法树定位到引起错误的操作
                offsets = None
                if frame_summary.lineno == frame_summary.end_lineno:
                    with suppress(Exception):
                        offsets = extract_caret_anchors_from_line_segment(code_segment)

                else:
                    # 不计算换行符，因为锚点只需定位到行的最后一个字符
                    end_offset = len(line.rstrip())

                # 如果主字符未跨越整行则添加定位锚行 ~~^~~
                if offsets and offsets[1] - offsets[0] > 0:
                    anchor_left, anchor_right = offsets

                    # 在终端显示时，某些非ASCII字符可能被渲染为双宽度字符，
                    # 因此在计算行长度时需要考虑这一点
                    dp_start_offset = display_width(line, start_offset) + 1
                    dp_left_end_offset = display_width(code_segment, anchor_left) + dp_start_offset
                    dp_right_start_offset = display_width(code_segment, anchor_right) + dp_start_offset
                    dp_end_offset = display_width(line, end_offset) + 1
                    anchor_indent = rc.translate("config.anchor.indent") + ' ' * (dp_start_offset - stripped_characters)
                    #  a =   1     /     0
                    # ....|~~~~~|^^^^^|~~~~~|
                    #   SO|  LEO|  RSO|   EO|
                    row.append(rc.anchors(
                        indent=anchor_indent,
                        left_start=dp_start_offset,
                        left_end=dp_left_end_offset,
                        right_start=dp_right_start_offset,
                        right_end=dp_end_offset,
                        crlf=True,
                    ))

                # 如果主字符未跨越整行则添加全指锚行 ^^^^^
                elif end_offset - start_offset < len(stripped_line):
                    dp_start_offset = display_width(line, start_offset) + 1
                    dp_end_offset = display_width(line, end_offset) + 1
                    anchor_indent = rc.translate("config.anchor.indent") + ' ' * (dp_start_offset - stripped_characters)
                    # what caaaaaan i say?
                    # ....|^^^^^^^^|
                    #   SO=LEO  RSO=EO
                    row.append(rc.anchors(
                        indent=anchor_indent,
                        left_start=dp_start_offset,
                        left_end=dp_start_offset,
                        right_start=dp_end_offset,
                        right_end=dp_end_offset,
                        crlf=True,
                    ))

        # 如果这个帧有携带的局部变量
        if frame_summary.locals_:
            # 逐个打印局部变量
            for name, value in sorted(frame_summary.locals_.items()):
                row.append(rc.translate('frame.location.locals_line', name=name, value=value))

        return ''.join(row)

    def format(self):
        """格式化堆栈信息以便打印。

        返回一个准备打印的字符串列表。结果列表中的每个字符串对应堆栈中的一个帧。
        每个字符串以换行符结尾；对于包含源代码文本行的项，字符串中可能也包含内部换行符。

        对于长时间重复的相同帧和行，会显示前几次重复，然后是一个总结行，说明确切的后续重复次数。
        """
        result = []
        last_file = None
        last_line = None
        last_name = None
        recursive_cutoff = rc.translate("config.stack.recursive_cutoff")
        if isinstance(recursive_cutoff, str) and recursive_cutoff.isdigit():
            recursive_cutoff = int(recursive_cutoff)
        if not isinstance(recursive_cutoff, int):
            import warnings
            warnings.warn(f"config.stack.recursive_cutoff must be an integer, but found ({type(recursive_cutoff)}){recursive_cutoff}, use default value {recursive_cutoff}")
            recursive_cutoff = 3

        count = 0
        for frame_summary in self:
            formatted_frame = self.format_frame_summary(frame_summary)
            if formatted_frame is None:
                continue

            # 重复组标记
            if (last_file is None or last_file != frame_summary.filename or
                    last_line is None or last_line != frame_summary.lineno or
                    last_name is None or last_name != frame_summary.name):  # 如果这是新的帧
                if count > recursive_cutoff:  # 且上一个帧重复了几次
                    count -= recursive_cutoff
                    result.append(rc.translate('config.stack.line_repeat_more', count=count))
                last_file = frame_summary.filename
                last_line = frame_summary.lineno
                last_name = frame_summary.name
                count = 0
            count += 1
            if count > recursive_cutoff:
                continue
            result.append(formatted_frame)

        if count > recursive_cutoff:
            count -= recursive_cutoff
            result.append(rc.translate('config.stack.line_repeat_more', count=count))
        return result



class KTBException:
    """
    一个准备好进行渲染的异常对象。

    回溯模块会从原始异常中捕获足够的属性到这个中间形式，以确保不保留任何引用，
    同时仍然能够完整地打印或格式化该异常。

    max_group_width 和 max_group_depth 控制异常组的格式化。
    深度指的是组的嵌套级别，
    宽度指的是单个异常组的异常数组的大小。
    当超过任一限制时，格式化输出将被截断。

    可以使用 `from_exception` 方法从异常对象创建 `TracebackException` 实例，
    也可以使用构造函数从单个组件创建 `TracebackException` 实例。

    - :attr:`__cause__` 原始异常 *__cause__* 对应的 `TracebackException` 实例。
    - :attr:`__context__` 原始异常 *__context__* 对应的 `TracebackException` 实例。
    - :attr:`exceptions` 对于异常组，这是一个包含嵌套 *exceptions* 对应的 `TracebackException` 实例的列表。对于其他异常，该值为 ``None``。
    - :attr:`__suppress_context__` 原始异常的 *__suppress_context__* 值。
    - :attr:`stack` 一个表示回溯信息的 `StackSummary` 对象。
    - :attr:`exc_type` 原始回溯的异常类。

    对于语法错误，还包含以下属性：
    - :attr:`filename` 语法错误发生的文件名。
    - :attr:`lineno` 语法错误发生的行号。
    - :attr:`end_lineno` 语法错误结束的行号。不存在时为 `None`。
    - :attr:`text` 语法错误发生处的文本。
    - :attr:`offset` 语法错误发生处文本的偏移量。
    - :attr:`end_offset` 语法错误结束处文本的偏移量。如果不存在，该值可以为 `None`。
    - :attr:`msg` 语法错误的编译器的错误消息。
    """
    _handler_types: list[Type["ErrorSuggestHandler"]] = []

    @classmethod
    def register(cls, Handler: Type["ErrorSuggestHandler"]):
        cls._handler_types.append(Handler)
        rc.register_handler(Handler)
        return Handler

    def __init__(self, exc_type, exc_value, exc_traceback, *, limit=None,
                 lookup_lines=True, capture_locals=False, compact=False,
                 max_group_width=15, max_group_depth=10, _seen=None):
        """
        初始化KTBException实例。

        参数:
        - exc_type: 异常类型
        - exc_value: 异常值
        - exc_traceback: 异常的回溯对象
        - limit: 限制回溯的深度
        - lookup_lines: 是否立即查找源代码行
        - capture_locals: 是否捕获局部变量
        - compact: 是否使用紧凑模式
        - max_group_width: 异常组的最大宽度
        - max_group_depth: 异常组的最大深度
        - _seen: 用于检测循环引用的已处理异常集合
        """


        # 处理 __cause__ 或 __context__ 中的循环。
        is_recursive_call = _seen is not None  # 是否是递归调用
        if _seen is None:
            _seen = set()
        _seen.add(id(exc_value))

        self.max_group_width = max_group_width
        self.max_group_depth = max_group_depth

        # 从回溯对象中提取堆栈信息
        self.stack: StackSummary = StackSummary.extract_from_extended_frame_gen(
            walk_tb_with_full_positions(exc_traceback),
            limit=limit, lookup_lines=lookup_lines,
            capture_locals=capture_locals)
        self.exc_type: Type[BaseException] = exc_type

        # 异常的字符串形式是它的原提示，比如"division by zero"
        # 或者'can only concatenate str (not "int") to str'
        self.exc_str = safe_string(exc_value, 'exception')

        # 获取异常的__notes__属性
        try:
            self.__notes__ = getattr(exc_value, '__notes__', None)
        except Exception as e:
            note_summary = safe_string(e, "__notes__", repr)
            self.__notes__ = [f'Ignored error getting __notes__: {note_summary}']

        self.final_exc_str = self.exc_str

        # 构建各个异常建议处理器便于后续提建议
        self._handlers: list["ErrorSuggestHandler"] = []
        for handler_type in self._handler_types:
            self._handlers.append(handler_type(exc_type, exc_value, exc_traceback, limit=limit,
                                               lookup_lines=lookup_lines, capture_locals=capture_locals,
                                               compact=compact, max_group_width=max_group_width,
                                               max_group_depth=max_group_depth, _seen=_seen))

        # 如果需要，加载源代码行
        if lookup_lines:
            self._load_lines()
        self.__suppress_context__ = \
            exc_value.__suppress_context__ if exc_value is not None else False

        # 使用队列处理__cause__和__context__，避免递归
        # 把所有的__cause__和__context__都挂到KTBException的__cause__, __context__, exceptions上, 完成引用解除
        self.exceptions: Optional[BaseExceptionGroup[BaseException]] = None
        if not is_recursive_call:  # 只允许首次创建的Kte创建处理队列，避免递归反复处理
            queue = [(self, exc_value)]
            while queue:
                te, e = queue.pop()
                # 处理cause
                if e and e.__cause__ is not None and id(e.__cause__) not in _seen:
                    cause = KTBException(
                        type(e.__cause__),
                        e.__cause__,
                        e.__cause__.__traceback__,
                        limit=limit,
                        lookup_lines=lookup_lines,
                        capture_locals=capture_locals,
                        max_group_width=max_group_width,
                        max_group_depth=max_group_depth,
                        _seen=_seen)
                else:
                    cause = None

                if compact:  # 紧凑模式下，只处理__suppress_context__为False的context
                    need_context = (cause is None and
                                    e is not None and
                                    not e.__suppress_context__)
                else:
                    need_context = True

                # 处理context
                if e and e.__context__ is not None and need_context and id(e.__context__) not in _seen:
                    context = KTBException(
                        type(e.__context__),
                        e.__context__,
                        e.__context__.__traceback__,
                        limit=limit,
                        lookup_lines=lookup_lines,
                        capture_locals=capture_locals,
                        max_group_width=max_group_width,
                        max_group_depth=max_group_depth,
                        _seen=_seen)
                else:
                    context = None

                # 处理异常组
                if e and isinstance(e, BaseExceptionGroup):
                    exceptions = []
                    for exc in e.exceptions:
                        texc = KTBException(
                            type(exc),
                            exc,
                            exc.__traceback__,
                            limit=limit,
                            lookup_lines=lookup_lines,
                            capture_locals=capture_locals,
                            max_group_width=max_group_width,
                            max_group_depth=max_group_depth,
                            _seen=_seen)
                        exceptions.append(texc)
                else:
                    exceptions = None

                te.__cause__ = cause
                te.__context__ = context
                te.exceptions = exceptions
                if cause:
                    queue.append((te.__cause__, e.__cause__))
                if context:
                    queue.append((te.__context__, e.__context__))
                if exceptions:
                    queue.extend(zip(te.exceptions, e.exceptions))

    @classmethod
    def from_exception(cls, exc, *args, **kwargs):
        """Create a TracebackException from an exception."""
        return cls(type(exc), exc, exc.__traceback__, *args, **kwargs)

    def _load_lines(self):
        """Private API. force all lines in the stack to be loaded."""
        for frame in self.stack:
            frame.load_line()

    def __eq__(self, other):
        if isinstance(other, KTBException):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __str__(self):
        return self.final_exc_str

    def format_exception_only(self):
        """格式化回溯中的异常部分。

        返回值是一个字符串Generator，每个字符串以换行符结尾。

        生成器会生成异常消息。

        对于 :exc:`SyntaxError` 异常，它还会在异常消息之前生成几行文本
        显示语法错误发生的详细位置信息。
        在异常消息之后，生成器还会生成该异常的所有 ``__notes__`` 属性内容。
        """
        if self.exc_type is None:
            if self.final_exc_str is None or not self.final_exc_str:
                yield rc.exc_line("UnknownError")
            else:
                yield rc.exc_line("UnknownError", self.final_exc_str)
            return

        stype: str = self.exc_type.__qualname__
        smod: str = self.exc_type.__module__
        if smod not in ("__main__", "builtins"):
            if not isinstance(smod, str):
                smod = "<unknown>"
            stype = smod + '.' + stype

        # 选择一个处理器，让它提提建议
        hi_priority = -1
        hi_priority_handler = None
        for handler in self._handlers:
            if handler.can_handle(self) and handler.priority > hi_priority:
                hi_priority = handler.priority
                hi_priority_handler = handler

        assert hi_priority_handler is not None, "No handler found for this exception. Why kwihandler.ErrorSuggestHandler not here?"
        yield from hi_priority_handler.handle(self)


        if (
            isinstance(self.__notes__, collections.abc.Sequence)
            and not isinstance(self.__notes__, (str, bytes))
        ):
            for note in self.__notes__:
                note = safe_string(note, 'note')
                yield from [l + '\n' for l in note.split('\n')]
        elif self.__notes__ is not None:
            yield "{}\n".format(safe_string(self.__notes__, '__notes__', func=repr))


    def format(self, *, chain =True, _ctx=None) -> Generator[str, None, None]:
        """格式化异常.

        如果 chain 不为 True，__cause__ 和 __context__ 不会递归地格式化。

        The return value is a generator of strings, each ending in a newline and
        some containing internal newlines. `print_exception` is a wrapper around
        this method which just prints the lines to a file.

        The message indicating which exception occurred is always the last
        string in the output.
        """

        if _ctx is None:
            _ctx = ExceptionPrintContext()

        output: list[tuple[Optional[str], Optional[KTBException]]] = []
        exc = self
        if chain:
            while exc:
                if exc.__cause__ is not None:
                    chained_msg = rc.translate("stack.cause")
                    chained_exc = exc.__cause__
                elif (exc.__context__ is not None and
                      not exc.__suppress_context__):
                    chained_msg = rc.translate("stack.context")
                    chained_exc = exc.__context__
                else:
                    chained_msg = None
                    chained_exc = None

                output.append((chained_msg, exc))
                exc = chained_exc
        else:
            output.append((None, exc))

        for msg, exc in reversed(output):
            if msg is not None:
                yield from _ctx.emit(msg)
            if exc.exceptions is None:
                if exc.stack:
                    yield from _ctx.emit(rc.translate("stack.summary"))
                    yield from _ctx.emit(exc.stack.format())
                yield from _ctx.emit(exc.format_exception_only())
            elif _ctx.exception_group_depth > self.max_group_depth:
                # 达到最大异常组深度，截断输出
                yield from _ctx.emit(
                    f"... (max_group_depth is {self.max_group_depth})\n")
            else:  # TODO: 这一整块是处理异常组的，但还没可翻译键值化，真烦人，异常组也没人用，以后再写
                # format exception group
                is_toplevel = (_ctx.exception_group_depth == 0)
                if is_toplevel:
                    _ctx.exception_group_depth += 1

                if exc.stack:
                    yield from _ctx.emit(
                        rc.translate("stack.group_summary"),
                        margin_char='+' if is_toplevel else None)
                    yield from _ctx.emit(exc.stack.format())

                yield from _ctx.emit(exc.format_exception_only())
                num_excs = len(exc.exceptions)
                if num_excs <= self.max_group_width:
                    n = num_excs
                else:
                    n = self.max_group_width + 1
                _ctx.need_close = False
                for i in range(n):
                    last_exc = (i == n - 1)
                    if last_exc:
                        # The closing frame may be added by a recursive call
                        _ctx.need_close = True

                    if self.max_group_width is not None:
                        truncated = (i >= self.max_group_width)
                    else:
                        truncated = False
                    title = f'{i + 1}' if not truncated else '...'
                    yield (_ctx.indent() +
                           ('+-' if i == 0 else '  ') +
                           f'+---------------- {title} ----------------\n')
                    _ctx.exception_group_depth += 1
                    if not truncated:
                        yield from exc.exceptions[i].format(chain=chain, _ctx=_ctx)
                    else:
                        remaining = num_excs - self.max_group_width
                        plural = 's' if remaining > 1 else ''
                        yield from _ctx.emit(
                            f"and {remaining} more exception{plural}\n")

                    if last_exc and _ctx.need_close:
                        yield (_ctx.indent() +
                               "+------------------------------------\n")
                        _ctx.need_close = False
                    _ctx.exception_group_depth -= 1

                if is_toplevel:
                    assert _ctx.exception_group_depth == 1
                    _ctx.exception_group_depth = 0

    def print(self, *, file=None, chain=True):
        """Print the result of self.format(chain=chain) to 'file'."""
        if file is None:
            file = sys.stderr
        for line in self.format(chain=chain):
            print(line, file=file, end="")




from typing import Generator
import inspect

import astroid

from kawaiitb.kraceback import KTBException, parse_module_filename
from kawaiitb.kwihandler import ErrorSuggestHandler
from kawaiitb.runtimeconfig import rc
from kawaiitb.utils import safe_string, is_sysstdlib_name
from kawaiitb.utils.suggestions import find_weighted_closest_matches, VarsGroup, merge_sorted_suggestions


@KTBException.register
class AttributeErrorHandler(ErrorSuggestHandler, priority=1.0):
    def __init__(self, exc_type, exc_value: AttributeError, exc_traceback, **kwargs):
        super().__init__(exc_type, exc_value, exc_traceback, **kwargs)
        self._can_handle = issubclass(exc_type, AttributeError)
        if not self._can_handle:
            return
        self.msg = safe_string(exc_value, "<exception>")
        obj = exc_value.obj
        self.obj = obj
        self.wrong_name = exc_value.name
        self.obj_rawname = safe_string(obj, "<unknown obj>")
        self.obj_type = safe_string(type(obj).__name__, "<unknown type>")
        self.obj_is_none = obj is None
        self.obj_is_module = False if obj is None else inspect.ismodule(obj)
        self.obj_module_name = getattr(obj, "__name__") if self.obj_is_module else None
        self.candidates = None
        try:
            self.candidates = dir(obj)
        except:
            pass

    def can_handle(self, etype: KTBException) -> bool:
        return self._can_handle

    @classmethod
    def translation_keys(cls):
        return {
            "default": {
                "native.AttributeError.premsg": "The '{obj}' of type {type_} has no attribute '{name}'.",
                "native.AttributeError.modulepremsg": "Module {obj_module_name} has no attribute '{name}'.",
                "native.AttributeError.circular_import": "This might be caused by a circular import.",
                "native.AttributeError.single_suggest": "Did you mean {suggest}?",
                "native.AttributeError.multiple_suggest": "Did you mean one of {suggest}?",
                "native.AttributeError.many_suggest": "Did you mean one of {suggest}?",
                "native.AttributeError.no_suggest": "",
                "native.AttributeError.list_all_property": "This object has the following attributes: {all}",
                "native.AttributeError.list_all_callable": "This object has the following methods: {all}",
                "native.AttributeError.no_any_prop": "This object has no attributes or methods!",
                "native.AttributeError.rename_from_shadowing_stdlib": "Module '{obj}' shadows a standard library module. '{attr}' exists in the corresponding standard library.\n",
                "native.AttributeError.rename_from_shadowing_stdlib_var": "Variable '{obj}' shadows a standard library module. '{attr}' exists in the corresponding standard library.\n",
                "native.AttributeError.interdependents_with_loop": "This is likely caused by a circular import, as the module has not completed initialization. Possible import cycle:\n{path}",
                "native.AttributeError.interdependents_no_info": "This is likely caused by a circular import, as the module has not completed initialization.",
                "native.AttributeError.interdependents_loop_table_top": "+->+",
                "native.AttributeError.interdependents_loop_table_mid": "| |",
                "native.AttributeError.interdependents_loop_table_low": "+<-+",
            },
            "zh_hans": {
                "native.AttributeError.premsg": "{type_} 类型的 '{obj}' 没有属性 '{name}'。",
                "native.AttributeError.modulepremsg": "模块 {obj_module_name} 没有属性 '{name}'。",
                "native.AttributeError.circular_import": "可能是循环导入导致的。",
                "native.AttributeError.single_suggest": "你可能想试试 {suggest}?",
                "native.AttributeError.multiple_suggest": "你可能想试试 {suggest}?",
                "native.AttributeError.many_suggest": "你可能想试试 {suggest}?",
                "native.AttributeError.no_suggest": "",
                "native.AttributeError.list_all_property": "该对象拥有以下属性: {all}{extra}",
                "native.AttributeError.list_all_callable": "该对象拥有以下方法: {all}{extra}",
                "native.AttributeError.no_any_prop": "该对象没有任何属性或方法！",
                "native.AttributeError.rename_from_shadowing_stdlib": "模块 '{obj}' 的名字覆盖了标准库模块。'{attr}' 存在于对应的标准库中。\n",
                "native.AttributeError.rename_from_shadowing_stdlib_var": "变量 '{obj}' 的名字覆盖了标准库模块。'{attr}' 存在于对应的标准库中。\n",
                "native.AttributeError.interdependents_with_loop": "这个问题很可能是循环导入导致的，因为该模块还完成初始化。可能的导入环如下: \n{path}",
                "native.AttributeError.interdependents_no_info": "这个问题很可能是循环导入导致的，因为该模块还未完成初始化。",
                "native.AttributeError.interdependents_loop_table_top": "┌→┐",
                "native.AttributeError.interdependents_loop_table_mid": "│ │",
                "native.AttributeError.interdependents_loop_table_low": "└←┘",
                "native.AttributeError.max_suggest_list": 20,  # 不建议改
            }
        }

    def handle(self, ktb_exc: KTBException) -> Generator[str, None, None]:
        yield from self._handle(ktb_exc)

    def _default_handle(self):
        if not self.msg:
            yield rc.exc_line("AttributeError", rc.translate("native.AttributeError.default"))
        else:
            yield rc.exc_line("AttributeError", self.msg)

    def get_interdependent_nodes(self, ktb_exc: KTBException):
        """
        基于可能的循环导入检查tb，如果找到循环环，返回循环链。否则返回None
        """
        if not self.obj_is_module or not ktb_exc or len(ktb_exc.stack) == 0:
            return None  # 不是模块/没有tb, 无法判断
        if not hasattr(self.obj, '__file__'):
            return None
        error_ocuring_file = getattr(self.obj, '__file__')
        start_recording = False
        interdependent_nodes = []
        for frame in ktb_exc.stack:
            if frame.filename == error_ocuring_file:
                start_recording = True
            if start_recording:
                interdependent_nodes.append(frame.filename)
        if len(interdependent_nodes) == 0:
            return None  # 不成环
        return interdependent_nodes  # 成环

    def _handle(self, ktb_exc: KTBException) -> Generator[str, None, None]:
        # 如果对象本身就有问题, 直接返回
        if not isinstance(self.wrong_name, str) or self.obj_is_none:
            yield from self._default_handle()
        if self.candidates is not None:
            candidates = self.candidates
        else:
            yield from self._default_handle()
            return

        # 获取代码当时使用时的情况
        if len(ktb_exc.stack) == 0:
            yield from self._default_handle()
            return

        # UC: 可调用对象(public, callable)
        # UP: 不可调用对象(public, property)
        # RC: 私有可调用对象(private, callable)
        # RP: 私有不可调用对象(private, property)
        # DU: 双下划线对象(double underscore)
        wrong_usage_type: VarsGroup
        # 是否私有
        wrong_usage_type = "R" if self.wrong_name.startswith('_') else "U"  # noqa
        # 是否可调用
        exc_frame = ktb_exc.stack[0]
        obj_rawname = self.obj_rawname
        attr_rawname = self.wrong_name
        for node in self.parse_ast_from_exc(exc_frame):
            if isinstance(node, (astroid.Import, astroid.ImportFrom)):
                extra_hint = ""
                if is_sysstdlib_name(self.obj_rawname) and self.obj_is_module:
                    extra_hint = rc.translate("native.AttributeError.rename_from_shadowing_stdlib")
                yield rc.exc_line("AttributeError", rc.translate("native.AttributeError.modulepremsg",
                                                                 obj_module_name=self.obj_module_name,
                                                                 name=attr_rawname) + extra_hint)
                break
            if isinstance(node, astroid.Attribute):
                if not isinstance(node.expr, (astroid.Name, astroid.Const)):  # 找出最终的不可进一步分解的节点
                    continue
                obj_rawname = node.expr.as_string()
                attr_rawname = node.attrname

                if is_sysstdlib_name(obj_rawname):
                    orig_lib = __import__(obj_rawname)
                    try:
                        getattr(orig_lib, attr_rawname)
                        if self.obj_is_module:
                            yield rc.translate('native.AttributeError.rename_from_shadowing_stdlib', obj=obj_rawname, attr=attr_rawname)
                        else:
                            yield rc.translate('native.AttributeError.rename_from_shadowing_stdlib_var', obj=obj_rawname, attr=attr_rawname)
                    except AttributeError:
                        pass


                if isinstance(node.parent, astroid.Call):
                    wrong_usage_type += "C"
                else:
                    wrong_usage_type += "P"
                break
        else:
            # never find an availd Attribute node
            yield from self._default_handle()
            return

        if "most likely due to a circular import" in str(ktb_exc):
            if loop := self.get_interdependent_nodes(ktb_exc):
                path = ""
                path_top = rc.translate("native.AttributeError.interdependents_loop_table_top")
                path_mid = rc.translate("native.AttributeError.interdependents_loop_table_mid")
                path_low = rc.translate("native.AttributeError.interdependents_loop_table_low")
                for i, filename in enumerate(loop):
                    path_table = path_top if i == 0 else path_mid if i < len(loop) - 1 else path_low
                    path += f"{path_table}{parse_module_filename(filename)}\n"
                yield rc.translate("native.AttributeError.interdependents_with_loop", obj=self.obj_module_name, attr=attr_rawname, path=path)
                return
            else:
                yield rc.translate("native.AttributeError.interdependents_no_info", obj=self.obj_module_name, attr=attr_rawname)
            return


        if self.wrong_name.startswith('__') and self.wrong_name.endswith('__'):
            wrong_usage_type = "DU"

        candidate_vars: dict[VarsGroup, list] = {"UC": [], "UP": [], "RC": [], "RP": [], "DU": []}
        for name in candidates:
            if not hasattr(self.obj, name):
                continue

            attr = getattr(self.obj, name)
            is_callable = callable(attr)
            is_private = name.startswith('_')
            is_dunder = name.startswith('__') and name.endswith('__')

            if is_dunder:
                candidate_vars["DU"].append(name)
            elif is_private:
                if is_callable:
                    candidate_vars["RC"].append(name)
                else:
                    candidate_vars["RP"].append(name)
            else:
                if is_callable:
                    candidate_vars["UC"].append(name)
                else:
                    candidate_vars["UP"].append(name)

        if wrong_usage_type != "DU" and not any((candidate_vars[t] for t in ("UC", "UP", "RC", "RP"))):  # noqa
            yield rc.translate("native.AttributeError.no_any_prop")
            return

        all_suggestions = find_weighted_closest_matches(self.wrong_name, candidate_vars)
        if all_suggestions[wrong_usage_type]:
            suggestions = [all_suggestions[wrong_usage_type][0]]
            if wrong_usage_type == "UP" and all_suggestions["UC"]:
                if all_suggestions["UP"][0][0] > all_suggestions["UC"][0][0]:
                    # print(f"{all1_suggestions["UP"][0][1]}={all_suggestions["UP"][0][0]}, {all_suggestions["UC"][0][1]}={all_suggestions["UC"][0][0]}")
                    suggestions.append(all_suggestions["UC"][0])
        else:
            suggestions = merge_sorted_suggestions(all_suggestions)[1]

        max_len = rc.translate("native.AttributeError.max_suggest_list")
        if len(suggestions) == 0:  # 没有针对性的建议，只能列出所有属性
            suggest_key = ("native.AttributeError.list_all_property", "native.AttributeError.list_all_callable") \
                [int("C" in wrong_usage_type)]
            suggesting = (
                candidate_vars["UC"] if wrong_usage_type == "UC" else
                candidate_vars["UP"] if wrong_usage_type == "UP" else
                candidate_vars["RC"] + candidate_vars["UC"] if wrong_usage_type == "RC" else
                candidate_vars["RP"] + candidate_vars["UP"] if wrong_usage_type == "RP" else
                candidate_vars["DU"]
            )
            extra = ', ...' if len(suggesting) > max_len else ''
            suggest_all = ', '.join(f"'{word}'" for word in suggesting[:max_len])

            suggestion_line = rc.translate("native.AttributeError.premsg",
                                           obj=obj_rawname,
                                           type_=self.obj_type,
                                           name=attr_rawname) + (
                rc.translate(suggest_key, all=suggest_all, extra=extra)
            )
        else:  # 有针对性的建议，即修改笔误
            extra = ', ...' if len(suggestions) > max_len else ''
            suggest_str = ', '.join(f"'{word}'" for _, word in suggestions[:max_len])
            suggestion_line = rc.translate("native.AttributeError.premsg",
                                           obj=obj_rawname,
                                           type_=self.obj_type,
                                           name=attr_rawname
                                           ) + (
                rc.translate("native.AttributeError.single_suggest", suggest=suggest_str) if len(suggestions) == 1 else
                rc.translate("native.AttributeError.multiple_suggest", suggest=suggest_str) if 1 < len(suggestions) <= 3 else
                rc.translate("native.AttributeError.many_suggest", suggest=suggest_str) if len(suggestions) > 3 else
                rc.translate("native.AttributeError.no_suggest")  # should never reach here
            )

        yield rc.exc_line("AttributeError", suggestion_line)



"""
This module incorporates functions adapted from the Python standard
library's traceback module, with modifications made to certain function
signatures. Due to significant variations in traceback.py across
different Python versions, it provides stable Python implementation
interfaces for consistent functionality implementation.

The source code references CPython 3.12 and adheres to the following
licensing terms:

Copyright (c) 2001-2024 Python Software Foundation.
All Rights Reserved.

Copyright (c) 2000 BeOpen.com.
All Rights Reserved.

Copyright (c) 1995-2001 Corporation for National Research Initiatives.
All Rights Reserved.

Copyright (c) 1991-1995 Stichting Mathematisch Centrum, Amsterdam.
All Rights Reserved.

Used under the PSF LICENSE AGREEMENT FOR PYTHON 3.12:
https://docs.python.org/3.12/license.html
"""
import itertools
import sys
import textwrap

__all__ = [
    "sentinel",
    "parse_value_tb",
    "walk_stack",
    "walk_tb",
    "walk_tb_with_full_positions",
    "get_code_position",
    "byte_offset_to_character_offset",
    "ExceptionPrintContext",
    "levenshtein_distance",
    "compute_suggestion_error",
    "substitution_cost",
    "display_width",
]


def walk_stack(f):
    """Walk a stack yielding the frame and line number for each frame.

    This will follow f.f_back from the given frame. If no frame is given, the
    current stack is used. Usually used with StackSummary.extract.
    """
    if f is None:
        f = sys._getframe().f_back.f_back.f_back.f_back
    while f is not None:
        yield f, f.f_lineno
        f = f.f_back


def walk_tb(tb):
    """Walk a traceback yielding the frame and line number for each frame.

    This will follow tb.tb_next (and thus is in the opposite order to
    walk_stack). Usually used with StackSummary.extract.
    """
    while tb is not None:
        yield tb.tb_frame, tb.tb_lineno
        tb = tb.tb_next


def walk_tb_with_full_positions(tb):
    # Internal version of walk_tb that yields full code positions including
    # end line and column information.
    while tb is not None:
        positions = get_code_position(tb.tb_frame.f_code, tb.tb_lasti)
        # Yield tb_lineno when co_positions does not have a line number to
        # maintain behavior with walk_tb.
        if positions[0] is None:
            yield tb.tb_frame, (tb.tb_lineno,) + positions[1:]
        else:
            yield tb.tb_frame, positions
        tb = tb.tb_next


def get_code_position(code, instruction_index):
    if instruction_index < 0:
        return None, None, None, None
    positions_gen = code.co_positions()
    return next(itertools.islice(positions_gen, instruction_index // 2, None))


def byte_offset_to_character_offset(str_, offset):
    as_utf8 = str_.encode('utf-8')
    return len(as_utf8[:offset].decode("utf-8", errors="replace"))


_WIDE_CHAR_SPECIFIERS = "WF"


def display_width(line, offset):
    """Calculate the extra amount of width space the given source
    code segment might take if it were to be displayed on a fixed
    width output device. Supports wide unicode characters and emojis."""

    # Fast track for ASCII-only strings
    if line.isascii():
        return offset

    import unicodedata

    return sum(
        2 if unicodedata.east_asian_width(char) in _WIDE_CHAR_SPECIFIERS else 1
        for char in line[:offset]
    )


class ExceptionPrintContext:
    def __init__(self):
        self.seen = set()
        self.exception_group_depth = 0
        self.need_close = False

    def indent(self):
        return ' ' * (2 * self.exception_group_depth)

    def emit(self, text_gen, margin_char=None):
        if margin_char is None:
            margin_char = '|'
        indent_str = self.indent()
        if self.exception_group_depth:
            indent_str += margin_char + ' '

        if isinstance(text_gen, str):
            yield textwrap.indent(text_gen, indent_str, lambda line: True)
        else:
            for text in text_gen:
                yield textwrap.indent(text, indent_str, lambda line: True)


_MAX_CANDIDATE_ITEMS = 750
_MAX_STRING_SIZE = 40
_MOVE_COST = 2
_CASE_COST = 1


def substitution_cost(ch_a, ch_b):
    if ch_a == ch_b:
        return 0
    if ch_a.lower() == ch_b.lower():
        return _CASE_COST
    return _MOVE_COST


def compute_suggestion_error(exc_value, tb, wrong_name):
    if wrong_name is None or not isinstance(wrong_name, str):
        return None
    if isinstance(exc_value, AttributeError):
        obj = exc_value.obj
        try:
            d = dir(obj)
        except Exception:
            return None
    elif isinstance(exc_value, ImportError):
        try:
            mod = __import__(exc_value.name)
            d = dir(mod)
        except Exception:
            return None
    else:
        assert isinstance(exc_value, NameError)
        # find most recent frame
        if tb is None:
            return None
        while tb.tb_next is not None:
            tb = tb.tb_next
        frame = tb.tb_frame
        d = (
                list(frame.f_locals)
                + list(frame.f_globals)
                + list(frame.f_builtins)
        )

        # Check first if we are in a method and the instance
        # has the wrong name as attribute
        if 'self' in frame.f_locals:
            self = frame.f_locals['self']
            if hasattr(self, wrong_name):
                return f"self.{wrong_name}"

    # Compute closest match

    if len(d) > _MAX_CANDIDATE_ITEMS:
        return None
    wrong_name_len = len(wrong_name)
    if wrong_name_len > _MAX_STRING_SIZE:
        return None
    best_distance = wrong_name_len
    suggestion = None
    for possible_name in d:
        if possible_name == wrong_name:
            # A missing attribute is "found". Don't suggest it (see GH-88821).
            continue
        # No more than 1/3 of the involved characters should need changed.
        max_distance = (len(possible_name) + wrong_name_len + 3) * _MOVE_COST // 6
        # Don't take matches we've already beaten.
        max_distance = min(max_distance, best_distance - 1)
        current_distance = levenshtein_distance(wrong_name, possible_name, max_distance)
        if current_distance > max_distance:
            continue
        if not suggestion or current_distance < best_distance:
            suggestion = possible_name
            best_distance = current_distance
    return suggestion


def levenshtein_distance(a, b, max_cost):
    # A Python implementation of Python/suggestions.c:levenshtein_distance.

    # Both strings are the same
    if a == b:
        return 0

    # Trim away common affixes
    pre = 0
    while a[pre:] and b[pre:] and a[pre] == b[pre]:
        pre += 1
    a = a[pre:]
    b = b[pre:]
    post = 0
    while a[:post or None] and b[:post or None] and a[post - 1] == b[post - 1]:
        post -= 1
    a = a[:post or None]
    b = b[:post or None]
    if not a or not b:
        return _MOVE_COST * (len(a) + len(b))
    if len(a) > _MAX_STRING_SIZE or len(b) > _MAX_STRING_SIZE:
        return max_cost + 1

    # Prefer shorter buffer
    if len(b) < len(a):
        a, b = b, a

    # Quick fail when a match is impossible
    if (len(b) - len(a)) * _MOVE_COST > max_cost:
        return max_cost + 1

    # Instead of producing the whole traditional len(a)-by-len(b)
    # matrix, we can update just one row in place.
    # Initialize the buffer row
    row = list(range(_MOVE_COST, _MOVE_COST * (len(a) + 1), _MOVE_COST))

    result = 0
    for bindex in range(len(b)):
        bchar = b[bindex]
        distance = result = bindex * _MOVE_COST
        minimum = sys.maxsize
        for index in range(len(a)):
            # 1) Previous distance in this row is cost(b[:b_index], a[:index])
            substitute = distance + substitution_cost(bchar, a[index])
            # 2) cost(b[:b_index], a[:index+1]) from previous row
            distance = row[index]
            # 3) existing result is cost(b[:b_index+1], a[index])

            insert_delete = min(result, distance) + _MOVE_COST
            result = min(insert_delete, substitute)

            # cost(b[:b_index+1], a[:index+1])
            row[index] = result
            if result < minimum:
                minimum = result
        if minimum > max_cost:
            # Everything in this row is too big, so bail early.
            return max_cost + 1
    return result


class _Sentinel:
    def __repr__(self):
        return "<implicit>"


sentinel = _Sentinel()


def parse_value_tb(exc, value, tb):
    if (value is sentinel) != (tb is sentinel):
        raise ValueError("Both or neither of value and tb must be given")
    if value is tb is sentinel:
        if exc is not None:
            if isinstance(exc, BaseException):
                return exc, exc.__traceback__

            raise TypeError(f'Exception expected for value, '
                            f'{type(exc).__name__} found')
        else:
            return None, None
    return value, tb



import sys
from io import TextIOWrapper, StringIO
from typing import Any, Callable, TextIO, Union

import kawaiitb.utils.fromtraceback as fromtraceback

sys_getframe = sys._getframe  # noqa

readables = (TextIO, TextIOWrapper, StringIO)
SupportsReading = Union[*readables]

__all__ = ["sys_getframe", "extract_caret_anchors_from_line_segment", "safe_string",
           "fromtraceback", "is_sysstdlib_name", *fromtraceback.__all__, ]


def safe_string(value: Any, what: str, func: Callable[[Any], str] = str):
    try:
        return func(value)
    except:
        return f'<{what} {func.__name__}() failed>'


def is_sysstdlib_name(name: str) -> bool:
    return name in sys.builtin_module_names or name in sys.stdlib_module_names


def extract_caret_anchors_from_line_segment(segment):
    import ast

    try:
        tree = ast.parse(segment)
    except SyntaxError:
        return None

    if len(tree.body) != 1:
        return None

    normalize = lambda offset: fromtraceback.byte_offset_to_character_offset(segment, offset)
    statement = tree.body[0]
    match statement:
        case ast.Expr(expr):
            match expr:
                case ast.BinOp():
                    operator_start = normalize(expr.left.end_col_offset)
                    operator_end = normalize(expr.right.col_offset)
                    operator_str = segment[operator_start:operator_end]
                    operator_offset = len(operator_str) - len(operator_str.lstrip())

                    left_anchor = expr.left.end_col_offset + operator_offset
                    right_anchor = left_anchor + 1
                    if operator_offset + 1 < len(operator_str) and not operator_str[operator_offset + 1].isspace():
                        right_anchor += 1

                    while left_anchor < len(segment) and ((ch := segment[left_anchor]).isspace() or ch in ")#"):
                        left_anchor += 1
                        right_anchor += 1
                    return normalize(left_anchor), normalize(right_anchor)
                case ast.Subscript():
                    left_anchor = normalize(expr.value.end_col_offset)
                    right_anchor = normalize(expr.slice.end_col_offset + 1)
                    while left_anchor < len(segment) and ((ch := segment[left_anchor]).isspace() or ch != "["):
                        left_anchor += 1
                    while right_anchor < len(segment) and ((ch := segment[right_anchor]).isspace() or ch != "]"):
                        right_anchor += 1
                    if right_anchor < len(segment):
                        right_anchor += 1
                    return left_anchor, right_anchor

    return None



"""
此模块管理本库的运行时配置。包括：
- 管理语言配置
使用 update_config 更新配置，更新配置之后使用 load_config 加载最新的配置。
- 翻译配置
load_config 加载好的配置可以借助rc(Runtime Config)使用。
使用 rc.set_language 设置语言。
使用 rc.translate 获取当前语言的翻译文本。
"""

__all__ = [
    "load_config",
    "update_config",
    "set_language",
    "rc",
]

import warnings
from collections import defaultdict, deque
from typing import TextIO, Type, TYPE_CHECKING, Optional

from kawaiitb._default_config import DEFAULT_CONFIG, EXTENDED
from kawaiitb.utils import readables

if TYPE_CHECKING:
    from kawaiitb.kwihandler import ErrorSuggestHandler

_config = DEFAULT_CONFIG.copy()


def update_config(config_data: dict):
    _config["translate_keys"] |= config_data["translate_keys"]
    _config["default_lang"] = config_data.get("default_lang", _config["default_lang"])


def load_config(config: dict | TextIO = None):
    """
    加载配置文件. 可以加载dict.
    """
    global _config
    if config is not None:
        if isinstance(config, readables):
            import json
            config_data = json.load(config)
        elif isinstance(config, dict):
            config_data = config
        else:
            raise TypeError("[KawaiiTB] config must be a dict or a file-like object")
        update_config(config_data)

    # 验证继承合法性
    for lang, data in _config["translate_keys"].items():
        if "extend" in data:
            extend_lang = data["extend"]
            if extend_lang not in _config["translate_keys"]:
                warnings.warn(f"[KawaiiTB] Language {extend_lang} is not supported, using default language instead.")
                _config["translate_keys"][lang]["extend"] = "default"
        elif lang != "default":
            _config["translate_keys"][lang]["extend"] = "default"

    # 验证非环 (Kahn法)
    adjacency = defaultdict(list)
    in_degree = defaultdict(int)
    for lang, data in _config["translate_keys"].items():
        if "extend" not in data:
            continue
        parent_lang = data["extend"]
        adjacency[parent_lang].append(lang)
        in_degree[lang] += 1
    queue = deque()
    for lang in _config["translate_keys"]:
        if in_degree[lang] == 0:
            queue.append(lang)
    processed = 0
    while queue:
        current = queue.popleft()
        processed += 1
        for neighbor in adjacency[current]:
            in_degree[neighbor] -= 1
            if in_degree[neighbor] == 0:
                queue.append(neighbor)
    if processed != len(_config["translate_keys"]):
        raise Exception("[KawaiiTB] A circular dependency exists in the language configuration!")

    # 重置默认语言
    if "rc" in globals():
        rc.change_language(_config["default_lang"])


def set_language(lang: str):
    rc.change_language(lang)


MAX_EXTEND_DEPTH = 10  # 最大继承深度


class RuntimeConfig:
    global _config

    def __init__(self):
        # 语言不一定非要是现存语言，也可以是额外的自定义提示信息(比如猫娘提示)
        self._lang = _config["default_lang"]
        self.langs = _config["translate_keys"].keys()
        self.default_config = _config["translate_keys"]["default"]
        self.handlers = []

    def register_handler(self, Handler: Type["ErrorSuggestHandler"]):
        """注册处理器的翻译键."""
        transkeys = Handler.translation_keys()
        if transkeys is not None and len(transkeys) > 0:
            for lang, data in transkeys.items():
                if lang not in _config["translate_keys"]:
                    _config["translate_keys"][lang] = {}
                _config["translate_keys"][lang] |= data
            load_config()

    def _get_key(self, lang, key):
        return _config["translate_keys"][lang][key]

    def _check_key(self, lang, key):
        return key in _config["translate_keys"][lang]

    def _check_lang(self, lang):
        return lang in self.langs

    def change_language(self, lang: str):
        if not self._check_lang(lang):
            warnings.warn(f"[KawaiiTB] Language {lang} is not exist, using default language instead.")
            self._lang = "default"
        else:
            self._lang = lang

    def translate(self, key: str, /, **kwargs):
        lang = self._lang
        depth_out = True
        for _ in range(MAX_EXTEND_DEPTH):
            if not self._check_lang(lang):  # 检查所选的语言是否存在, 否则使用默认语言
                warnings.warn(f"[KawaiiTB] Language {lang} is not exist, using default language instead.")
                warnings.warn('langs:' + str(self.langs))
                lang = "default"
                continue

            if self._check_key(lang, key) \
                    and self._get_key(lang, key) != EXTENDED:  # 优先使用该语言已有的配置
                if kwargs:
                    return self._get_key(lang, key).format(**kwargs)
                return self._get_key(lang, key)

            if lang == "default":  # 如果连默认配置都不存在这个键，返回未知配置提示
                warnings.warn(f"[KawaiiTB] Unknown translate key: {key}")
                return f"<Unknown config: {key}>"

            if self._check_key(lang, "extend"):  # 如果继承自其他语言……
                lang = self._get_key(lang, "extend")
                continue

            # 其他情况，走默认配置
            lang = "default"

        warnings.warn(f"[KawaiiTB] Exceeded maximum extend depth for key: {key}")
        return f"<[KawaiiTB] Unreachable translate key: {key}>"

    def anchors(self, indent: str, left_start: int, left_end: int, right_start: int, right_end: int, crlf: bool):
        primary = self.translate("config.anchor.primary")
        secondary = self.translate("config.anchor.secondary")

        left_segment = primary * (left_end - left_start)
        middle_segment = secondary * (right_start - left_end)
        right_segment = primary * (right_end - right_start)
        suffix = self.translate("config.anchor.suffix")
        if crlf:
            suffix += "\n"

        return indent + left_segment + middle_segment + right_segment + suffix

    def exc_line(self, exc_type_str: str, exc_value_str: Optional[str]=None):
        if exc_value_str is None:
            return self.translate("exception.exc_line_noval", etype=exc_type_str)
        else:
            return self.translate("exception.exc_line", etype=exc_type_str, value=exc_value_str)

rc: RuntimeConfig = RuntimeConfig()



import sys
from io import StringIO

import pytest

from test.utils.utils import KTBTestBase

custom_conf = """{
  "translate_keys": {
    "test_conf": {
      "stack.summary": "Test summary qwertyuiopasdfghjklzxcvbnm\\n"
    },
    "another_lang": {
      "stack.summary": "Test summary 1234567890\\n"
    }
  },
  "default_lang": "test_conf"
}"""  # "\\n" is necessary
test_conf = "Test summary qwertyuiopasdfghjklzxcvbnm"
another_lang = "Test summary 1234567890"
kirakira = "☆pypy被玩坏了☆这肯定不是py的问题☆绝对不是☆"


class TestKTBLoad(KTBTestBase, console_output=False):
    def test_module_load(self):
        def is_changed():
            return sys.excepthook != sys.__excepthook__

        assert not is_changed(), "默认的异常钩子未被正确加载"

        import kawaiitb
        assert not is_changed(), "import kawaiitb 时异常钩子被提前加载"

        kawaiitb.load()
        assert is_changed(), "kawaiitb.load() 未能正确加载异常钩子"

        kawaiitb.unload()
        assert not is_changed(), "kawaiitb.unload() 未能正确卸载异常钩子"

        import kawaiitb.autoload  # noqa
        assert is_changed(), "import kawaiitb.load 未能正确自动加载异常钩子"

    def get_exception_summary(self):
        from kawaiitb import kraceback
        with pytest.raises(ZeroDivisionError) as excinfo:
            _ = 1 / 0
        e = excinfo.value
        self.try_print_exc(e)
        return "".join(kraceback.format_exception(e))

    def test_load_defualt(self):
        from kawaiitb import load, unload
        load()
        assert "KawaiiTB" in self.get_exception_summary()
        unload()

    def test_load_lang(self):
        from kawaiitb import load, unload
        load('kirakira')
        assert kirakira in self.get_exception_summary()
        unload()

    def test_load_fileio(self):
        from kawaiitb import load, unload
        load(StringIO(custom_conf))
        assert test_conf in self.get_exception_summary()
        unload()

    def test_load_fileio_lang(self):
        from kawaiitb import load, unload
        load('another_lang', StringIO(custom_conf))
        assert another_lang in self.get_exception_summary()
        unload()
        assert sys.excepthook == sys.__excepthook__, "kawaiitb.unload() 未能正确卸载异常钩子"



r"""
kawaii-traceback 多语言配置系统说明

===== 静态扩展翻译键 =====

语言配置是一个json文件或python字典，包含以下结构:
{
    "translate_keys": 翻译键值对
        "(your_lang)": 你的翻译名
            "extend": str;你要继承的翻译名  <== 如果没有，默认继承default
            "(some.keywords)": str;各种翻译键  <== 可翻译键值对可以在下面的default中找到
    "default_lang": str;默认语言  <== 如果希望你的翻译加载后能够立刻生效，在这个键中指定你的翻译名
}
注意：
- 翻译名应当是有效的python标识符 + "(要继承的翻译名)". 继承的翻译名可以省略其父继承
- 通过包含 extend 键指定父配置. 实际的继承关系以此键名决定. 未指定会自动继承default.
- 在python字典中使用EXTENDED显示指定一个翻译键的父翻译. 在json中使用"==EXTEND==".

===== 动态扩展翻译键 =====

你可以通过继承 kawaiitb.ErrorSuggestHandler来定义一个动态扩展的**提示信息**.
提示信息只在一部分动态生成。例如：

...... Traceback (most recent call last):
......   File "C:\Users\BPuffer\Desktop\kawaii-traceback\main.py", line 74:8, in <module>
......     asyncio.run(asyncio.sleep(1))
......     ^^^^^^^
HINT-> NameError: name 'asyncio' is not defined
HINT-> 猜你没导入: 'asyncio'

动态生成的灵活性很高，你除了可以获取异常本身外还能获得当前运行代码的上下文信息。比如使用本库还原的原生语法错误特殊解析：
...... Traceback (most recent call last):
......   File "C:\Users\BPuffer\Desktop\kawaii-traceback\main.py", line 138:8, in <module>
......     exec("what can i say?")
HINT->   File "<string>", line 1
HINT->     what can i say?
HINT->          ^^^
HINT-> SyntaxError: invalid syntax (<string>, line 1)

---

如果你希望使用动态扩展，参阅 kawaiitb.kwihandler.py 中的示例.
"""
__all__ = [
    "DEFAULT_CONFIG",
    "EXTENDED",
]

EXTENDED = "==EXTEND=="  # 标记继承父翻译项

DEFAULT_CONFIG = {
    "translate_keys": {
        "default": {
            # 交互式提示符
            "config.prompt1": ">>> ",
            "config.prompt2": "... ",

            # 重复的帧
            "config.stack.recursive_cutoff": 3,
            "config.stack.line_repeat_more": '  [Previous line repeated {count} more times]\n',

            # 文件路径解析
            "config.file.include_cwd": True,
            "config.file.parse_module_filename": False,
            "config.file.parsed_filename": "[{namespace}] {filename}",
            "config.module": "<module>",
            "config.string": "<string>",

            # 锚点
            "config.anchor.indent": ' ' * 4,
            "config.anchor.primary": '~',
            "config.anchor.secondary": '^',
            "config.anchor.suffix": '',

            # 单帧格式化
            "frame.location.with_column": '  File "{file}", line {lineno}:{colno}, in {name}\n',
            "frame.location.without_column": '  File "{file}", line {lineno}, in {name}\n',
            "frame.location.without_name": '  File "{file}", line {lineno}\n',
            "frame.location.linetext": '    {line}\n',
            "frame.location.locals_line": '    {name} = {value}\n',

            # 帧栈格式化
            "stack.cause": "\nThe above exception was the direct cause of the following exception:\n\n",
            "stack.context": "\nDuring handling of the above exception, another exception occurred:\n\n",
            "stack.summary": "Traceback (most recent call last):\n",
            "stack.group_summary": "Exception Group Traceback (most recent call last):\n",

            # 各个异常格式化
            "exception.message": "{etype}: {value}\n",
            "exception.exc_line_noval": "{etype}\n",
            "exception.exc_line": "{etype}: {value}\n",
            "exception.qualname": "{qualname}",
            "exception.module_qualname": "{module}.{qualname}",
            "exception.note": "{note}\n",
        },
        "en_us": {},  # 继承default
        "zh_hans": {  # 简体中文配置
            "config.file.include_cwd": False,
            "config.module": "模块级语句",
            "config.string": "注入的语句",
            "frame.location.with_column": '  文件 {file}:{lineno}:{colno} 的 {name}\n',
            "frame.location.without_column": '  文件 {file}:{lineno} 的 {name}\n',
            "frame.location.without_name": '  文件 "{file}:{lineno}"\n',
            "stack.cause": "\n该异常引发了另一个异常:\n\n",
            "stack.context": "\n处理上面的异常时，发生了如下异常:\n\n",
            "stack.summary": "异常回溯 (到最近一次调用):\n",
            "config.stack.line_repeat_more": ' (*在那之后, 这一行重复了 {count} 次)\n',
        },
        "neko_zh": {  # 萌化中文配置示例
            "extend": "zh_hans",
            "stack.summary": "pypy被玩坏了！这肯定不是py的问题！绝对不是！\n",
            "config.file.parse_module_filename": True,
            "config.file.parsed_filename": "[{namespace} 模块] {filename}",
            "config.prompt1": "owo!> ",
            "config.prompt2": "=w=~| ",
            "config.anchor.suffix": " ↖在这里喵~",

            "exception.message": "[{etype}] {value}\n",
            "exception.exc_line_noval": "[{etype}]!\n",
            "exception.exc_line": "[{etype}] {value}\n",
        },
        "kirakira": {  # 二次扩展示例
            "extend": "neko_zh",
            "stack.summary": "☆pypy被玩坏了☆这肯定不是py的问题☆绝对不是☆\n",
            "config.prompt1": "✧⋆˚｡~˚~｡> ",
            "config.prompt2": "✧ :*✧･ﾟ:| ",
        }
    },
    "default_lang": "neko_zh",
}

if __name__ == "__main__":
    # 示例自定义配置
    new_config = {
        "translate_keys": {
            "neko_zh": {
                "extend": "zh_hans",
                "config.file.parse_module_filename": True,
                "frame.location.with_column": EXTENDED,
                "config.anchor.suffix": " ↖在这里喵~",
            }
        },
        "default_lang": "neko_zh"
    }
    # 此处以字典格式为例。JSON配置可以使用`kawaiitb.load(open('your_config.json'))`



import os
import sys
from functools import wraps
from traceback import format_exception as orig_format_exception
from typing import overload, Optional, Literal

from kawaiitb.kraceback import KTBException
from kawaiitb.runtimeconfig import rc, load_config
from kawaiitb.utils.fromtraceback import parse_value_tb, sentinel as _sentinel
from kawaiitb.utils import SupportsReading, readables

ENV_KAWAIITB_LANG = "KAWAIITB_LANG"
ENV_KAWAIITB_CONF = "KAWAIITB_CONF"

__excepthook__ = sys.excepthook
__in_console__ = hasattr(sys, 'ps1')
if __in_console__:
    __ps1__ = sys.ps1
    __ps2__ = sys.ps2


def unload():
    sys.excepthook = __excepthook__
    if __in_console__:
        sys.ps1 = __ps1__
        sys.ps2 = __ps2__


@overload
def load(*, excepthook: bool = True, console_prompt: bool = True):
    """加载默认配置"""


@overload
def load(lang: str = None, *, excepthook: bool = True, console_prompt: bool = True):
    """选取一个语言"""


@overload
def load(file: SupportsReading = None, *, excepthook: bool = True, console_prompt: bool = True,
         encoding: Optional[str] = 'utf-8'):
    """加载一个配置文件"""


@overload
def load(lang: Literal[None] = None, filename: str = None, *, excepthook: bool = True, console_prompt: bool = True,
         encoding: Optional[str] = 'utf-8'):
    """从文件名加载一个配置文件"""


@overload
def load(lang: str = None, file: SupportsReading | str = None, *, excepthook: bool = True, console_prompt: bool = True,
         encoding: Optional[str] = 'utf-8'):
    """加载一个配置文件并选择一个语言"""


def load(lang: Optional[SupportsReading | str] = None,
         file: Optional[SupportsReading | str] = None,
         *,
         excepthook: bool = True,
         console_prompt: bool = True,
         encoding: Optional[str] = 'utf-8',
         ):
    """
    加载配置并劫持hook.

    :param lang: 要加载的语言. 如果为None, 则加载默认配置.
    :param file: 配置文件. 如果为None, 则加载默认配置.
    :param excepthook: 是否劫持sys.excepthook. 默认值为True.
    :param console_prompt: 是否劫持控制台提示符. 默认值为True.
    :param encoding: 配置文件的编码. 默认值为'utf-8'.

    usage:
>>> import kawaiitb
>>> kawaiitb.load()  # 加载默认配置。读取环境变量KAWAIITB_LANG和KAWAIITB_CONF
>>> kawaiitb.load('neko_zh')  # 加载指定语言的配置
>>> kawaiitb.load(file='my_conf.json')  # 加载指定文件的配置
>>> kawaiitb.load('present1', 'hot_conf.json')  # 加载指定语言和文件的配置
>>> with open('my_conf.json') as f:
...     load(f)  # 加载指定文件的配置
>>> with open('hot_conf.json') as f:
...     load('present1', f)  # 加载指定语言和文件的配置
    """
    if isinstance(lang, readables) and file is None:
        lang, file = None, lang  # When call with a single file-like object, same as load(file=...)
    if not isinstance(lang, (str, type(None))):
        raise ValueError(f"Invalid type of lang {type(lang)}")
    if not isinstance(file, (str, type(None), *readables)):
        raise ValueError(f"Invalid type of file {type(file)}")

    # 默认情况: load()
    if lang is None and file is None:
        load_config()
        if ENV_KAWAIITB_LANG in os.environ:
            rc.change_language(os.environ[ENV_KAWAIITB_LANG])

        if ENV_KAWAIITB_CONF in os.environ:
            with open(os.environ[ENV_KAWAIITB_CONF], 'r', encoding=encoding) as f:
                load_config(f)
        override(excepthook=excepthook, console_prompt=console_prompt)
        return

    # 用法: load('neko_zh')
    if isinstance(lang, str) and file is None:
        load_config()
        rc.change_language(lang)
        override(excepthook=excepthook, console_prompt=console_prompt)
        return

    # 用法: load(file='my_conf.json')
    if lang is None and isinstance(file, str):
        with open(file, 'r', encoding=encoding) as f:
            load_config(f)
        override(excepthook=excepthook, console_prompt=console_prompt)
        return

    # 用法: load('present1', 'hot_conf.json')
    if isinstance(lang, str) and isinstance(file, str):
        with open(file, 'r', encoding=encoding) as f:
            load_config(f)
        rc.change_language(lang)
        override(excepthook=excepthook, console_prompt=console_prompt)
        return

    # 用法: with open('my_conf.json') as f: load(f)
    if lang is None and hasattr(file, 'readable') and file.readable():
        load_config(file)
        override(excepthook=excepthook, console_prompt=console_prompt)
        return

    # 用法: with open('hot_conf.json') as f: load(f, 'present1')
    if isinstance(lang, str) and hasattr(file, 'readable') and file.readable():
        load_config(file)
        rc.change_language(lang)
        override(excepthook=excepthook, console_prompt=console_prompt)
        return

    raise ValueError("Invalid arguments")


def override(excepthook=True, console_prompt=None):
    if excepthook:
        @wraps(orig_format_exception)  # 签名对齐 traceback.format_exception
        def wrapped(exc, /, value=_sentinel, tb=_sentinel, limit=None, chain=True):
            try:
                value, tb = parse_value_tb(exc, value, tb)
                te = KTBException(type(value), value, tb, limit=limit, compact=True)
                return list(te.format(chain=chain))
            except Exception as ktb_self_raised_exc:
                return orig_format_exception(exc, value=_sentinel, tb=_sentinel, limit=None, chain=True) + [
                    "\n\nKawaiiTB occurred another exception while formatting this exception:\n"
                ] + list(orig_format_exception(ktb_self_raised_exc)) + [
                    "Please report this to the KawaiiTB developers.\n"
                ]

        wrapped.__kawaiitb__ = True  # take over by KawaiiTB
        sys.excepthook = wrapped

    if (
            console_prompt is True and
            __in_console__ and
            (ps1 := rc.translate("config.prompt1")) and
            (ps2 := rc.translate("config.prompt2"))
    ):
        sys.ps1 = ps1
        sys.ps2 = ps2



"""
This file is used to load KawaiiTB with a single command.
Usage:
>>> import kawaiitb.autoload
"""

from kawaiitb.tools import load, ENV_KAWAIITB_LANG, ENV_KAWAIITB_CONF
import os
if not os.environ.get(ENV_KAWAIITB_LANG) and not os.environ.get(ENV_KAWAIITB_CONF):
    load('neko_zh')  # 喵~
else:
    load()



import astroid


def is_point_before(line: int, col: int, start_line: int, start_col: int):
    return (
            (line < start_line) or
            (line == start_line and col < start_col)
    )


def is_point_after(line: int, col: int, end_line: int, end_col: int):
    return (
            (line > end_line) or
            (line == end_line and col > end_col)
    )


def is_point_inside(line: int, col: int, start_line: int, end_line: int, start_col: int, end_col: int):
    if start_line == line == end_line:
        return start_col <= col <= end_col
    return (
            (line == start_line and col >= start_col) or
            (line == end_line and col <= end_col) or
            (start_line < line < end_line)
    )


def is_completely_inside(node: astroid.nodes.NodeNG, start_line: int, end_line: int, start_col: int, end_col: int):
    if any((
        not hasattr(node, 'lineno') or node.lineno is None,
        not hasattr(node, 'col_offset') or node.col_offset is None,
        not hasattr(node, 'end_lineno') or node.end_lineno is None,
        not hasattr(node, 'end_col_offset') or node.end_col_offset is None,
    )): return False
    return (
            is_point_inside(node.lineno, node.col_offset, start_line, end_line, start_col, end_col) and
            is_point_inside(node.end_lineno, node.end_col_offset, start_line, end_line, start_col, end_col)
    )


def is_partially_inside(node: astroid.nodes.NodeNG, start_line: int, end_line: int, start_col: int, end_col: int):
    if any((
        not hasattr(node, 'lineno') or node.lineno is None,
        not hasattr(node, 'col_offset') or node.col_offset is None,
        not hasattr(node, 'end_lineno') or node.end_lineno is None,
        not hasattr(node, 'end_col_offset') or node.end_col_offset is None,
    )): return False
    return (
            is_point_before(node.lineno, node.col_offset, end_line, end_col) or
            is_point_after(node.end_lineno, node.end_col_offset, start_line, start_col)
    )


def astroid_walk_inside(node: astroid.nodes.NodeNG, start_line: int, end_line: int, start_col: int, end_col: int):
    if is_completely_inside(node, start_line, end_line, start_col, end_col):
        yield node
    for child in node.get_children():  # 这里不能检查部分包含，因为部分大节点可能缺结束位，无法判断是否包含
        if isinstance(child, astroid.nodes.NodeNG) and \
                is_partially_inside(child, start_line, end_line, start_col, end_col):
            yield from astroid_walk_inside(child, start_line, end_line, start_col, end_col)


